"""routines for automatic calibration.

- hpge_find_E_peaks (Find uncalibrated E peaks whose E spacing matches the pattern in peaks_keV)
- hpge_get_E_peaks (Get uncalibrated E peaks at the energies of peaks_keV)
- hpge_fit_E_peaks (fits the energy peals)
- hpge_E_calibration (main routine -- finds and fits peaks specified)
"""
import logging
import sys

import matplotlib.gridspec as gs
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
from iminuit import Minuit, cost
from scipy.signal import find_peaks_cwt, medfilt

import pygama.math.histogram as pgh
import pygama.math.peak_fitting as pgf
import pygama.math.utils as pgu
from pygama.pargen.utils import return_nans

log = logging.getLogger(__name__)


def hpge_find_E_peaks(
    hist,
    bins,
    var,
    peaks_keV,
    n_sigma=5,
    deg=0,
    Etol_keV=None,
    var_zero=1,
    verbose=False,
):
    """Find uncalibrated E peaks whose E spacing matches the pattern in peaks_keV
    Note: the specialization here to units "keV" in peaks and Etol is
    unnecessary. However it is kept so that the default value for Etol_keV has
    an unambiguous interpretation.

    Parameters
    ----------
    hist, bins, var : array, array, array
        Histogram of uncalibrated energies, see pgh.get_hist()
        var cannot contain any zero entries.
    peaks_keV : array
        Energies of peaks to search for (in keV)
    n_sigma : float
        Threshold for detecting a peak in sigma (i.e. sqrt(var))
    deg : int
        deg arg to pass to poly_match
    Etol_keV : float
        absolute tolerance in energy for matching peaks
    var_zero : float
        number used to replace zeros of var to avoid divide-by-zero in
        hist/sqrt(var). Default value is 1. Usually when var = 0 its because
        hist = 0, and any value here is fine.
    verbose : bool
        print debug messages

    Returns
    -------
    detected_peak_locations : list
        list of uncalibrated energies of detected peaks
    detected_peak_energies : list
        list of calibrated energies of detected peaks
    pars : list of floats
        the parameters for poly(peaks_uncal) = peaks_keV (polyfit convention)
    """
    # clean up var if necessary
    if np.any(var == 0):
        log.debug(f"hpge_find_E_peaks: replacing var zeros with {var_zero}")
        var[np.where(var == 0)] = var_zero
    peaks_keV = np.asarray(peaks_keV)

    # Find all maxes with > n_sigma significance
    imaxes = get_i_local_maxima(hist / np.sqrt(var), n_sigma)

    # Now pattern match to peaks_keV within Etol_keV using poly_match
    detected_max_locs = pgh.get_bin_centers(bins)[imaxes]

    if Etol_keV is None:
        # estimate Etol_keV
        pt_pars, pt_covs = hpge_fit_E_peak_tops(
            hist, bins, var, detected_max_locs, n_to_fit=15
        )
        if (
            sum(sum(sum(c) if (c != None).any() else 0 for c in pt_covs)) == np.inf
            or sum(sum(sum(c) if (c != None).any() else 0 for c in pt_covs)) == 0
        ):
            log.debug(
                "hpge_find_E_peaks: can safely ignore previous covariance warning, not used"
            )
        pt_pars = pt_pars[np.array([x is not None for x in pt_pars])]
        med_sigma_ratio = np.median(np.stack(pt_pars)[:, 1] / np.stack(pt_pars)[:, 0])

        Etol_keV = 5.0 * (med_sigma_ratio / 0.003)
    pars, ixtup, iytup = poly_match(
        detected_max_locs, peaks_keV, deg=deg, atol=Etol_keV
    )

    if len(ixtup) != len(peaks_keV):
        log.info(
            f"hpge_find_E_peaks: only found {len(ixtup)} of {len(peaks_keV)} expected peaks"
        )
    return detected_max_locs[ixtup], peaks_keV[iytup], pars


def hpge_get_E_peaks(
    hist,
    bins,
    var,
    cal_pars,
    peaks_keV,
    n_sigma=3,
    Etol_keV=5,
    var_zero=1,
    verbose=False,
):
    """Get uncalibrated E peaks at the energies of peaks_keV

    Parameters
    ----------
    hist, bins, var : array, array, array
        Histogram of uncalibrated energies, see pgh.get_hist()
        var cannot contain any zero entries.
    cal_pars : array
        Estimated energy calibration parameters used to search for peaks
    peaks_keV : array
        Energies of peaks to search for (in keV)
    n_sigma : float
        Threshold for detecting a peak in sigma (i.e. sqrt(var))
    Etol_keV : float
        absolute tolerance in energy for matching peaks
    var_zero : float
        number used to replace zeros of var to avoid divide-by-zero in
        hist/sqrt(var). Default value is 1. Usually when var = 0 its because
        hist = 0, and any value here is fine.
    verbose : bool
        print debug messages

    Returns
    -------
    got_peak_locations : list
        list of uncalibrated energies of found peaks
    got_peak_energies : list
        list of calibrated energies of found peaks
    pars : list of floats
        the parameters for poly(peaks_uncal) = peaks_keV (polyfit convention)
    """
    # clean up var if necessary
    if np.any(var == 0):
        log.debug(f"hpge_find_E_peaks: replacing var zeros with {var_zero}")
        var[np.where(var == 0)] = var_zero
    peaks_keV = np.asarray(peaks_keV)

    # Find all maxes with > n_sigma significance
    imaxes = get_i_local_maxima(hist / np.sqrt(var), n_sigma)

    # Keep maxes if they coincide with expected peaks
    test_peaks_keV = np.asarray([pgf.poly(i, cal_pars) for i in bins[imaxes]])
    imatch = [abs(peaks_keV - i).min() < Etol_keV for i in test_peaks_keV]

    got_peak_locations = bins[imaxes[imatch]]
    got_peak_energies = test_peaks_keV[imatch]

    # Match calculated and true peak energies
    matched_energies = peaks_keV[
        [np.argmin(abs(peaks_keV - i)) for i in got_peak_energies]
    ]
    while not all([list(matched_energies).count(x) == 1 for x in matched_energies]):
        for i in range(len(matched_energies)):
            if matched_energies[i + 1] == matched_energies[i]:
                # remove duplicates
                if np.argmin(
                    abs(got_peak_energies[i : i + 2] - matched_energies[i])
                ):  # i+1 is best match
                    got_peak_locations = np.delete(got_peak_locations, i)
                    got_peak_energies = np.delete(got_peak_energies, i)
                else:  # i is best match
                    got_peak_locations = np.delete(got_peak_locations, i + 1)
                    got_peak_energies = np.delete(got_peak_energies, i + 1)
                matched_energies = np.delete(matched_energies, i)
                break
            i += 1

    # Calculate updated calibration curve
    pars = np.polyfit(got_peak_locations, matched_energies, len(cal_pars))

    return got_peak_locations, matched_energies, pars


def hpge_fit_E_peak_tops(
    hist,
    bins,
    var,
    peak_locs,
    n_to_fit=7,
    cost_func="Least Squares",
    inflate_errors=False,
    gof_method="var",
):
    """Fit gaussians to the tops of peaks

    Parameters
    ----------
    hist, bins, var : array, array, array
        Histogram of uncalibrated energies, see pgh.get_hist()
    peak_locs : array
        locations of peaks in hist. Must be accurate two within +/- 2*n_to_fit
    n_to_fit : int
        number of hist bins near the peak top to include in the gaussian fit
    cost_func : bool (optional)
        Flag passed to gauss_mode_width_max()
    inflate_errors : bool (optional)
        Flag passed to gauss_mode_width_max()
    gof_method : str (optional)
        method flag passed to gauss_mode_width_max()

    Returns
    -------
    pars_list : list of array
        a list of best-fit parameters (mode, sigma, max) for each peak-top fit
    cov_list : list of 2D arrays
        a list of covariance matrices for each pars
    """
    pars_list = []
    cov_list = []
    for E_peak in peak_locs:
        try:
            pars, cov = pgf.gauss_mode_width_max(
                hist,
                bins,
                var,
                mode_guess=E_peak,
                n_bins=n_to_fit,
                cost_func=cost_func,
                inflate_errors=inflate_errors,
                gof_method=gof_method,
            )
        except:
            pars, cov = None, None

        pars_list.append(pars)
        cov_list.append(cov)
    return np.array(pars_list, dtype=object), np.array(cov_list, dtype=object)


def get_hpge_E_peak_par_guess(hist, bins, var, func, mode_guess):
    """Get parameter guesses for func fit to peak in hist

    Parameters
    ----------
    hist, bins, var : array, array, array
        Histogram of uncalibrated energies, see pgh.get_hist(). Should be
        windowed around the peak.
    func : function
        The function to be fit to the peak in the (windowed) hist
    """
    if (
        func == pgf.gauss_step_cdf
        or func == pgf.gauss_step_pdf
        or func == pgf.extended_gauss_step_pdf
    ):
        # get mu and height from a gauss fit, also sigma as fallback
        pars, cov = pgf.gauss_mode_width_max(
            hist, bins, var, mode_guess=mode_guess, n_bins=10
        )
        bin_centres = pgh.get_bin_centers(bins)
        if pars is None:
            log.info("get_hpge_E_peak_par_guess: gauss_mode_width_max failed")
            i_0 = np.argmax(hist)
            mu = bin_centres[i_0]
            height = hist[i_0]
            sigma_guess = None
        else:
            mu = mode_guess
            sigma_guess = pars[1]
            height = pars[2]

        # get bg and step from edges of hist
        bg = np.mean(hist[-10:])
        step = bg - np.mean(hist[:10])
        # get sigma from fwfm with f = 1/sqrt(e)
        try:
            sigma = pgh.get_fwfm(
                0.6065,
                hist,
                bins,
                var,
                mx=height,
                bl=bg - step / 2,
                method="interpolate",
            )[0]
            if sigma == 0:
                raise ValueError
        except:
            sigma = pgh.get_fwfm(
                0.6065,
                hist,
                bins,
                var,
                mx=height,
                bl=bg - step / 2,
                method="fit_slopes",
            )[0]
            if sigma == 0:
                log.info("get_hpge_E_peak_par_guess: sigma estimation failed")
                if sigma_guess is not None:
                    sigma = sigma_guess
                else:
                    return []

        # now compute amp and return
        n_sig = np.sum(
            hist[(bin_centres > mu - 3 * sigma) & (bin_centres < mu + 3 * sigma)]
        )
        n_bkg = np.sum(hist) - n_sig

        hstep = step / (bg + np.mean(hist[:10]))

        parguess = [n_sig, mu, sigma / 2, n_bkg, hstep, bins[0], bins[-1], 0]
        for i, guess in enumerate(parguess):
            if np.isnan(guess):
                parguess[i] = 0

        return parguess

    if (
        func == pgf.radford_cdf
        or func == pgf.radford_pdf
        or func == pgf.extended_radford_pdf
    ):
        # guess mu, height
        pars, cov = pgf.gauss_mode_width_max(
            hist, bins, var, mode_guess=mode_guess, n_bins=10
        )
        bin_centres = pgh.get_bin_centers(bins)
        if pars is None:
            log.info("get_hpge_E_peak_par_guess: gauss_mode_width_max failed")
            sigma_guess = None

        else:
            sigma_guess = pars[1]
            # mu=pars[0]
            # height=pars[2]
        i_0 = np.argmax(hist)
        mu = bin_centres[i_0]
        height = hist[i_0]

        # get bg and step from edges of hist
        bg0 = np.mean(hist[-10:])
        step = bg0 - np.mean(hist[:10])

        # get sigma from fwfm with f = 1/sqrt(e)
        try:
            sigma = pgh.get_fwfm(
                0.6065,
                hist,
                bins,
                var,
                mx=height,
                bl=bg0 + step / 2,
                method="interpolate",
            )[0]
            if sigma == 0:
                raise ValueError
        except:
            sigma = pgh.get_fwfm(
                0.6065,
                hist,
                bins,
                var,
                mx=height,
                bl=bg0 + step / 2,
                method="fit_slopes",
            )[0]
            if sigma == 0:
                log.info("get_hpge_E_peak_par_guess: sigma estimation failed")
                if sigma_guess is not None:
                    sigma = sigma_guess
                else:
                    return []
        sigma = sigma * 0.8  # roughly remove some amount due to tail

        # for now hard-coded
        htail = 1.0 / 5
        tau = 0.5 * sigma

        hstep = step / (bg0 + np.mean(hist[:10]))

        n_sig = np.sum(
            hist[(bin_centres > mu - 3 * sigma) & (bin_centres < mu + 3 * sigma)]
        )
        n_bkg = np.sum(hist) - n_sig

        parguess = [n_sig, mu, sigma, htail, tau, n_bkg, hstep, bins[0], bins[-1], 0]

        for i, guess in enumerate(parguess):
            if np.isnan(guess):
                parguess[i] = 0

        return parguess

    else:
        log.error(f"get_hpge_E_peak_par_guess not implemented for {func.__name__}")
        return []


def get_hpge_E_fixed(func):
    """
    Returns: Sequence list of fixed indexes for fitting and mask for parameters
    """

    if (
        func == pgf.gauss_step_cdf
        or func == pgf.gauss_step_pdf
        or func == pgf.extended_gauss_step_pdf
    ):
        # pars are: n_sig, mu, sigma, n_bkg, hstep, components
        return [5, 6, 7], np.array([True, True, True, True, True, False, False, False])

    if (
        func == pgf.radford_cdf
        or func == pgf.radford_pdf
        or func == pgf.extended_radford_pdf
    ):
        # pars are: n_sig, mu, sigma, htail,tau, n_bkg, hstep, components
        return [7, 8, 9], np.array(
            [True, True, True, True, True, True, True, False, False, False]
        )

    else:
        log.error(f"get_hpge_E_fixed not implemented for {func.__name__}")
        return None
    return None


def get_hpge_E_bounds(func, parguess):
    if (
        func == pgf.radford_cdf
        or func == pgf.radford_pdf
        or func == pgf.extended_radford_pdf
    ):
        return [
            (0, None),
            (parguess[-3], parguess[-2]),
            (0, None),
            (0, 1),
            (None, None),
            (0, None),
            (-1, 1),
            (None, None),
            (None, None),
            (None, None),
        ]

    elif (
        func == pgf.gauss_step_cdf
        or func == pgf.gauss_step_pdf
        or func == pgf.extended_gauss_step_pdf
    ):
        return [
            (0, None),
            (parguess[-3], parguess[-2]),
            (0, None),
            (0, None),
            (-1, 1),
            (None, None),
            (None, None),
            (None, None),
        ]

    else:
        log.error(f"get_hpge_E_bounds not implemented for {func.__name__}")
        return []


class tail_prior:
    """
    Generic least-squares cost function with error.
    """

    verbose = 0
    errordef = Minuit.LIKELIHOOD  # for Minuit to compute errors correctly

    def __init__(self, data, model, tail_weight=100):
        self.model = model  # model predicts y for given x
        self.data = data
        self.tail_weight = tail_weight

    def _call(self, *pars):
        return self.__call__(*pars[0])

    def __call__(
        self,
        n_sig,
        mu,
        sigma,
        htail,
        tau,
        n_bkg,
        hstep,
        lower_range,
        upper_range,
        components,
    ):
        return self.tail_weight * np.log(htail + 0.1)  # len(self.data)/


def staged_fit(
    energies, hist, bins, var, func_i, gof_func_i, simplex, mode_guess, tail_weight=100
):
    par_guesses = get_hpge_E_peak_par_guess(hist, bins, var, func_i, mode_guess)
    bounds = get_hpge_E_bounds(func_i, par_guesses)
    fixed, mask = get_hpge_E_fixed(func_i)

    if func_i == pgf.extended_radford_pdf or func_i == pgf.radford_pdf:
        cost_func = cost.ExtendedUnbinnedNLL(energies, func_i) + tail_prior(
            energies, func_i, tail_weight=tail_weight
        )
        m = Minuit(cost_func, *par_guesses)
        m.limits = bounds
        for fix in fixed:
            m.fixed[fix] = True

        m.values["htail"] = 0
        m.values["tau"] = 0
        m.fixed["htail"] = True
        m.fixed["tau"] = True
        if simplex == True:
            m.simplex().migrad()
        else:
            m.migrad()
        try:
            # set htail to guess
            m.values["htail"] = par_guesses[3]
            m.values["tau"] = par_guesses[4]
            m.fixed = False
            for fix in fixed:
                m.fixed[fix] = True

            if simplex == True:
                m.simplex().migrad()
            else:
                m.migrad()
            m.hesse()
            pars_i = m.values
            errs_i = m.errors
            cov_i = m.covariance
            valid_fit = m.valid
            if valid_fit == False:
                raise RuntimeError
        except:
            func_i = pgf.extended_gauss_step_pdf
            gof_func_i = pgf.gauss_step_pdf
            pars_i, errs_i, cov_i, func_i, gof_func_i, mask, valid_fit = staged_fit(
                energies, hist, bins, var, func_i, gof_func_i, simplex, mode_guess
            )

        # check htail
        if (
            m.values["htail"] < 0.01
            or m.values["htail"] < 2 * m.errors["htail"]
            or np.isnan(m.values).any()
        ):  # switch to stat test
            func_i = pgf.extended_gauss_step_pdf
            gof_func_i = pgf.gauss_step_pdf
            pars_i, errs_i, cov_i, func_i, gof_func_i, mask, valid_fit = staged_fit(
                energies, hist, bins, var, func_i, gof_func_i, simplex, mode_guess
            )

    else:
        cost_func = cost.ExtendedUnbinnedNLL(energies, func_i)
        m = Minuit(cost_func, *par_guesses)
        m.limits = bounds
        for fix in fixed:
            m.fixed[fix] = True
        if simplex == True:
            m.simplex().migrad()
        else:
            m.migrad()

        m.hesse()

        pars_i = m.values
        errs_i = m.errors
        cov_i = m.covariance

        valid_fit = m.valid

    return pars_i, errs_i, cov_i, func_i, gof_func_i, mask, valid_fit


def hpge_fit_E_peaks(
    E_uncal,
    mode_guesses,
    wwidths,
    n_bins=50,
    funcs=pgf.gauss_step_cdf,
    method="unbinned",
    gof_funcs=None,
    n_events=None,
    allowed_p_val=0.05,
    uncal_is_int=False,
    simplex=False,
    tail_weight=100,
):
    """Fit the Energy peaks specified using the function given

    Parameters
    ----------
    E_uncal : array
        unbinned energy data to be fit
    mode_guesses : array
        array of guesses for modes of each peak
    wwidths : float or array of float
        array of widths to use for the fit windows (in units of E_uncal),
        typically on the order of 10 sigma where sigma is the peak width
    n_bins : int or array of ints
        array of number of bins to use for the fit window histogramming
    funcs : function or array of functions
        funcs to be used to fit each region
    method : str
        default is unbinned fit can specify to use binned fit method instead
    gof_funcs : function or array of functions
        functions to use for calculation goodness of fit if unspecified will use same func as fit
    uncal_is_int : bool
        if True, attempts will be made to avoid picket-fencing when binning
        E_uncal
    simplex : bool determining whether to do a round of simpson minimisation before gradient minimisation
    n_events : int number of events to use for unbinned fit
    allowed_p_val : lower limit on p_val of fit

    Returns
    -------
    pars : list of array
        a list of best-fit parameters for each peak fit
    covs : list of 2D arrays
        a list of covariance matrices for each pars
    binwidths : list
        a list of bin widths used for each peak fit
    ranges: list of array
        a list of [Euc_min, Euc_max] used for each peak fit
    """
    pars = np.zeros(len(mode_guesses), dtype="object")
    errors = np.zeros(len(mode_guesses), dtype="object")
    covs = np.zeros(len(mode_guesses), dtype="object")
    binws = np.zeros(len(mode_guesses))
    ranges = np.zeros(len(mode_guesses), dtype="object")
    p_vals = np.zeros(len(mode_guesses))
    valid_pks = np.zeros(len(mode_guesses), dtype=bool)
    out_funcs = np.zeros(len(mode_guesses), dtype="object")

    for i_peak, mode_guess in enumerate(mode_guesses):
        # get args for this peak
        wwidth_i = wwidths if not isinstance(wwidths, list) else wwidths[i_peak]
        n_bins_i = n_bins if np.isscalar(n_bins) else n_bins[i_peak]
        func_i = funcs[i_peak] if hasattr(funcs, "__len__") else funcs
        wleft_i = wwidth_i / 2 if np.isscalar(wwidth_i) else wwidth_i[0]
        wright_i = wwidth_i / 2 if np.isscalar(wwidth_i) else wwidth_i[1]
        if gof_funcs is not None:
            gof_func_i = (
                gof_funcs[i_peak] if hasattr(gof_funcs, "__len__") else gof_funcs
            )
        else:
            gof_func_i = func_i

        try:
            # bin a histogram
            Euc_min = mode_guesses[i_peak] - wleft_i
            Euc_max = mode_guesses[i_peak] + wright_i
            if uncal_is_int == True:
                Euc_min, Euc_max, n_bins_i = pgh.better_int_binning(
                    x_lo=Euc_min, x_hi=Euc_max, n_bins=n_bins_i
                )

            if method == "unbinned":
                energies = E_uncal[(E_uncal > Euc_min) & (E_uncal < Euc_max)][:n_events]
                hist, bins, var = pgh.get_hist(
                    energies, bins=n_bins_i, range=(Euc_min, Euc_max)
                )
                if func_i == pgf.extended_radford_pdf or pgf.extended_gauss_step_pdf:
                    (
                        pars_i,
                        errs_i,
                        cov_i,
                        func_i,
                        gof_func_i,
                        mask,
                        valid_fit,
                    ) = staged_fit(
                        energies,
                        hist,
                        bins,
                        var,
                        func_i,
                        gof_func_i,
                        simplex,
                        mode_guess,
                        tail_weight=tail_weight,
                    )
                    if pars_i["n_sig"] < 20:
                        valid_fit = False
                else:
                    par_guesses = get_hpge_E_peak_par_guess(hist, bins, var, func_i)
                    bounds = get_hpge_E_bounds(func_i, par_guesses)
                    fixed, mask = get_hpge_E_fixed(func_i)

                    cost_func = cost.ExtendedUnbinnedNLL(energies, func_i)
                    m = Minuit(cost_func, *par_guesses)
                    m.limits = bounds
                    for fix in fixed:
                        m.fixed[fix] = True
                    if simplex == True:
                        m.simplex().migrad()
                    else:
                        m.migrad()
                    m.hesse()

                    pars_i = m.values
                    errs_i = m.errors
                    cov_i = m.covariance
                    valid_fit = m.valid

                csqr = pgf.goodness_of_fit(
                    hist,
                    bins,
                    None,
                    gof_func_i,
                    pars_i,
                    method="Pearson",
                    scale_bins=True,
                )

            else:
                hist, bins, var = pgh.get_hist(
                    E_uncal, bins=n_bins_i, range=(Euc_min, Euc_max)
                )
                par_guesses = get_hpge_E_peak_par_guess(hist, bins, var, func_i)
                bounds = get_hpge_E_bounds(func_i, par_guesses)
                fixed, mask = get_hpge_E_fixed(func_i)
                pars_i, errs_i, cov_i = pgf.fit_binned(
                    func_i,
                    hist,
                    bins,
                    var=var,
                    guess=par_guesses,
                    cost_func=method,
                    Extended=True,
                    fixed=fixed,
                    simplex=simplex,
                    bounds=bounds,
                )
                valid_fit = True

                csqr = pgf.goodness_of_fit(
                    hist,
                    bins,
                    None,
                    gof_func_i,
                    pars_i,
                    method="Pearson",
                    scale_bins=False,
                )

            if np.isnan(pars_i).any():
                log.debug(
                    f"hpge_fit_E_peaks: fit failed for i_peak={i_peak} at loc {mode_guesses[i_peak]:g}, par is nan : {pars_i}"
                )
                raise RuntimeError

            p_val = scipy.stats.chi2.sf(csqr[0], csqr[1] + len(np.where(mask)[0]))

            total_events = pgf.get_total_events_func(func_i, pars_i, errors=errs_i)
            if (
                sum(sum(c) if c is not None else 0 for c in cov_i[mask, :][:, mask])
                == np.inf
                or sum(sum(c) if c is not None else 0 for c in cov_i[mask, :][:, mask])
                == 0
                or np.isnan(
                    sum(sum(c) if c is not None else 0 for c in cov_i[mask, :][:, mask])
                )
            ):
                log.debug(
                    f"hpge_fit_E_peaks: cov estimation failed for i_peak={i_peak} at loc {mode_guesses[i_peak]:g}"
                )
                valid_pks[i_peak] = False

            elif valid_fit == False:
                log.debug(
                    f"hpge_fit_E_peaks: peak fitting failed for i_peak={i_peak} at loc {mode_guesses[i_peak]:g}"
                )
                valid_pks[i_peak] = False

            elif (
                np.abs(np.array(errs_i)[mask] / np.array(pars_i)[mask]) < 1e-7
            ).any() or np.isnan(np.array(errs_i)[mask]).any():
                log.debug(
                    f"hpge_fit_E_peaks: failed for i_peak={i_peak} at loc {mode_guesses[i_peak]:g}, parameter error too low"
                )
                valid_pks[i_peak] = False

            elif np.abs(total_events[0] - np.sum(hist)) / np.sum(hist) > 0.1:
                log.debug(
                    f"hpge_fit_E_peaks: fit failed for i_peak={i_peak} at loc {mode_guesses[i_peak]:g}, total_events is outside limit"
                )
                valid_pks[i_peak] = False

            elif p_val < allowed_p_val or np.isnan(p_val):
                log.debug(
                    f"hpge_fit_E_peaks: fit failed for i_peak={i_peak}, p-value too low: {p_val}"
                )
                valid_pks[i_peak] = False
            else:
                valid_pks[i_peak] = True

        except:
            log.debug(
                f"hpge_fit_E_peaks: fit failed for i_peak={i_peak}, unknown error"
            )
            valid_pks[i_peak] = False
            pars_i, errs_i, cov_i = return_nans(func_i)
            p_val = 0

        # get binning
        binw_1 = (bins[-1] - bins[0]) / (len(bins) - 1)

        pars[i_peak] = pars_i
        errors[i_peak] = errs_i
        covs[i_peak] = cov_i
        binws[i_peak] = binw_1
        ranges[i_peak] = [Euc_min, Euc_max]
        p_vals[i_peak] = p_val
        out_funcs[i_peak] = func_i

    return (pars, errors, covs, binws, ranges, p_vals, valid_pks, out_funcs)


def poly_wrapper(x, *pars):
    return pgf.poly(x, pars)


def hpge_fit_E_scale(mus, mu_vars, Es_keV, deg=0):
    """Find best fit of poly(E) = mus +/- sqrt(mu_vars)
    Compare to hpge_fit_E_cal_func which fits for E = poly(mu)

    Parameters
    ----------
    mus : array
        uncalibrated energies
    mu_vars : array
        variances in the mus
    Es_keV : array
        energies to fit to, in keV
    deg : int
        degree for energy scale fit. deg=0 corresponds to a simple scaling
        mu = scale * E. Otherwise deg follows the definition in np.polyfit

    Returns
    -------
    pars : array
        parameters of the best fit. Follows the convention in np.polyfit
    cov : 2D array
        covariance matrix for the best fit parameters.
    """
    if deg == 0:
        scale, scale_cov = pgu.fit_simple_scaling(Es_keV, mus, var=mu_vars)
        pars = np.array([scale, 0])
        cov = np.array([[scale_cov, 0], [0, 0]])
        errs = np.diag(np.sqrt(cov))
    else:
        poly_pars = np.polyfit(Es_keV, mus, deg=deg, w=1 / np.sqrt(mu_vars))
        c = cost.LeastSquares(Es_keV, mus, np.sqrt(mu_vars), poly_wrapper)
        m = Minuit(c, *poly_pars)
        m.simplex()
        m.migrad()
        m.hesse()
        pars = m.values
        cov = m.covariance
        errs = m.errors
    return pars, errs, cov


def hpge_fit_E_cal_func(mus, mu_vars, Es_keV, E_scale_pars, deg=0):
    """Find best fit of E = poly(mus +/- sqrt(mu_vars))
    This is an inversion of hpge_fit_E_scale.
    E uncertainties are computed from mu_vars / dmu/dE where mu = poly(E) is the
    E_scale function

    Parameters
    ----------
    mus : array
        uncalibrated energies
    mu_vars : array
        variances in the mus
    Es_keV : array
        energies to fit to, in keV
    E_scale_pars : array
        Parameters from the escale fit (keV to ADC) used for calculating
        uncertainties
    deg : int
        degree for energy scale fit. deg=0 corresponds to a simple scaling
        mu = scale * E. Otherwise deg follows the definition in np.polyfit

    Returns
    -------
    pars : array
        parameters of the best fit. Follows the convention in np.polyfit
    cov : 2D array
        covariance matrix for the best fit parameters.
    """
    if deg == 0:
        E_vars = mu_vars / E_scale_pars[0] ** 2
        scale, scale_cov = pgu.fit_simple_scaling(mus, Es_keV, var=E_vars)
        pars = np.array([scale, 0])
        cov = np.array([[scale_cov, 0], [0, 0]])
        errs = np.diag(np.sqrt(cov))
    else:
        dmudEs = np.zeros(len(mus))
        for n in range(len(E_scale_pars) - 1):
            dmudEs += E_scale_pars[n] * mus ** (len(E_scale_pars) - 2 - n)
        E_weights = dmudEs * mu_vars
        poly_pars = np.polyfit(mus, Es_keV, deg=deg, w=1 / E_weights)
        c = cost.LeastSquares(mus, Es_keV, E_weights, poly_wrapper)
        m = Minuit(c, *poly_pars)
        m.simplex()
        m.migrad()
        m.hesse()
        pars = m.values
        errs = m.errors
        cov = m.covariance
    return pars, errs, cov


def hpge_E_calibration(
    E_uncal,
    peaks_keV,
    guess_keV,
    deg=0,
    uncal_is_int=False,
    range_keV=None,
    funcs=pgf.gauss_step_cdf,
    gof_funcs=None,
    method="unbinned",
    gof_func=None,
    n_events=None,
    simplex=False,
    allowed_p_val=0.05,
    tail_weight=100,
    verbose=True,
):
    """Calibrate HPGe data to a set of known peaks

    Parameters
    ----------
    E_uncal : array
        unbinned energy data to be calibrated
    peaks_keV : array
        list of peak energies to be fit to. Each must be in the data
    guess_keV : float
        a rough initial guess at the conversion factor from E_uncal to keV. Must
        be positive
    deg : non-negative int
        degree of the polynomial for the E_cal function E_keV = poly(E_uncal).
        deg = 0 corresponds to a simple scaling E_keV = scale * E_uncal.
        Otherwise follows the convention in np.polyfit
    uncal_is_int : bool
        if True, attempts will be made to avoid picket-fencing when binning
        E_uncal
    range_keV : float, tuple, array of floats, or array of tuples of floats
        ranges around which the peak fitting is performed
        if tuple(s) are supplied, they provide the left and right ranges
    funcs
        DOCME
    gof_funcs : function or array of functions
        functions to use for calculation goodness of fit if unspecified will use same func as fit
    method : str
        default is unbinned fit can specify to use binned fit method instead
    gof_func
        DOCME
    n_events : int
        number of events to use for unbinned fit
    simplex : bool
        DOCME
    allowed_p_val
        lower limit on p_val of fit
    verbose : bool
        print debug statements

    Returns
    -------
    pars, cov : array, 2D array
        array of calibration function parameters and their covariances. The form
        of the function is E_keV = poly(E_uncal). Assumes poly() is
        overwhelmingly dominated by the linear term. pars follows convention in
        np.polyfit unless deg=0, in which case it is the (lone) scale factor
    results : dict with the following elements
        'detected_peaks_locs', 'detected_peaks_keV' : array, array
            array of rough uncalibrated/calibrated energies at which the fit peaks were
            found in the initial peak search
        'pt_pars', 'pt_cov' : list of (array), list of (2D array)
            arrays of gaussian parameters / covariances fit to the peak tops in
            the first refinement
        'pt_cal_pars', 'pt_cal_cov' : array, 2D array
            array of calibration parameters E_uncal = poly(E_keV) for fit to
            means of gausses fit to tops of each peak
        'pk_pars', 'pk_cov', 'pk_binws', 'pk_ranges' : list of (array), list of (2D array), list, list of (array)
            the best fit parameters, covariances, bin width and energy range for the local fit to each peak
        'pk_cal_pars', 'pk_cal_cov' : array, 2D array
            array of calibration parameters E_uncal = poly(E_keV) for fit to
            means from full peak fits
        'fwhms', 'dfwhms' : array, array
            the numeric fwhms and their uncertainties for each peak.
    """
    results = {}

    if not isinstance(range_keV, list):
        range_keV = [range_keV for peak in peaks_keV]

    if not hasattr(funcs, "__len__"):
        funcs = [funcs for peak in peaks_keV]

    # sanity checks
    E_uncal = np.asarray(E_uncal)
    peaks_keV = np.asarray(peaks_keV)  # peaks_keV = np.sort(peaks_keV)
    deg = int(deg)
    if guess_keV <= 0:
        log.error(f"hpge_E_cal warning: invalid guess_keV = {guess_keV}")
        return None, None, results
    if deg < 0:
        log.error(f"hpge_E_cal warning: invalid deg = {deg}")
        return None, None, results

    # bin the histogram in ~1 keV bins for the initial rough peak search
    Euc_min = peaks_keV[0] / guess_keV * 0.6
    Euc_max = peaks_keV[-1] / guess_keV * 1.1
    dEuc = 1 / guess_keV
    if uncal_is_int:
        Euc_min, Euc_max, dEuc = pgh.better_int_binning(
            x_lo=Euc_min, x_hi=Euc_max, dx=dEuc
        )
    hist, bins, var = pgh.get_hist(E_uncal, range=(Euc_min, Euc_max), dx=dEuc)

    # Run the initial rough peak search
    detected_peaks_locs, detected_peaks_keV, roughpars = hpge_find_E_peaks(
        hist, bins, var, peaks_keV, n_sigma=5, deg=deg
    )
    log.info(f"{len(detected_peaks_locs)} peaks found:")
    log.info(f"\t   Energy   | Position  ")
    for i, (Li, Ei) in enumerate(zip(detected_peaks_locs, detected_peaks_keV)):
        log.info(f"\t{i}".ljust(4) + str(Ei).ljust(9) + f"| {Li:g}".ljust(5))

    # re-bin the histogram in ~0.2 keV bins with updated E scale par for peak-top fits
    Euc_min, Euc_max = (
        (np.poly1d(roughpars) - i).roots
        for i in (peaks_keV[0] * 0.9, peaks_keV[-1] * 1.1)
    )
    Euc_min = Euc_min[np.logical_and(Euc_min >= 0, Euc_min <= max(Euc_max))][0]
    Euc_max = Euc_max[
        np.logical_and(Euc_max >= Euc_min, Euc_max <= np.nanmax(E_uncal) * 1.1)
    ][0]
    dEuc = 0.2 / roughpars[-2]

    if uncal_is_int:
        Euc_min, Euc_max, dEuc = pgh.better_int_binning(
            x_lo=Euc_min, x_hi=Euc_max, dx=dEuc
        )
    hist, bins, var = pgh.get_hist(E_uncal, range=(Euc_min, Euc_max), dx=dEuc)

    # run peak getter after rebinning
    got_peaks_locs, got_peaks_keV, roughpars = hpge_get_E_peaks(
        hist, bins, var, roughpars, peaks_keV, n_sigma=3
    )
    results["got_peaks_locs"] = got_peaks_locs
    results["got_peaks_keV"] = got_peaks_keV

    log.info(f"{len(got_peaks_locs)} peaks obtained:")
    log.info(f"\t   Energy   | Position  ")
    for i, (Li, Ei) in enumerate(zip(got_peaks_locs, got_peaks_keV)):
        log.info(f"\t{i}".ljust(4) + str(Ei).ljust(9) + f"| {Li:g}".ljust(5))

    # Drop non-gotten peaks
    idx = [i for i, E in enumerate(peaks_keV) if E in got_peaks_keV]
    range_keV = [range_keV[i] for i in idx]
    funcs = [funcs[i] for i in idx]
    gof_funcs = [gof_funcs[i] for i in idx]

    # Drop peaks to not be fitted
    tmp = zip(
        *[
            a
            for a in zip(got_peaks_locs, got_peaks_keV, range_keV, funcs)
            if a[2] and a[3]
        ]
    )
    got_peaks_locs, got_peaks_keV, range_keV, funcs = list(map(list, tmp))
    got_peaks_locs = np.asarray(got_peaks_locs)
    got_peaks_keV = np.asarray(got_peaks_keV)

    # Now do a series of full fits to the peak shapes

    # First calculate range around peaks to fit
    if range_keV is None:
        # Need to do initial fit
        pt_pars, pt_covs = hpge_fit_E_peak_tops(
            hist, bins, var, got_peaks_locs, n_to_fit=7
        )
        # Drop failed fits
        fitidx = [i is not None for i in pt_pars]
        results["got_peaks_locs"] = got_peaks_locs = got_peaks_locs[fitidx]
        results["got_peaks_keV"] = got_peaks_keV = got_peaks_keV[fitidx]
        pt_pars = np.asarray(pt_pars)[fitidx]
        pt_covs = np.asarray(pt_covs)[fitidx]
        range_uncal = np.stack(pt_pars)[:, 1].astype(float) * 20
        n_bins = 50
    elif np.isscalar(range_keV):
        derco = np.polyder(np.poly1d(roughpars)).coefficients
        der = [pgf.poly(Ei, derco) for Ei in got_peaks_keV]
        range_uncal = [float(range_keV) / d for d in der]
        n_bins = [int(range_keV / 0.5 / d) for d in der]
    elif isinstance(range_keV, tuple):
        rangeleft_keV, rangeright_keV = range_keV
        derco = np.polyder(np.poly1d(roughpars)).coefficients
        der = [pgf.poly(Ei, derco) for Ei in got_peaks_keV]
        range_uncal = [(rangeleft_keV / d, rangeright_keV / d) for d in der]
        n_bins = [int(sum(range_keV) / 0.5 / d) for d in der]
    elif isinstance(range_keV, list):
        derco = np.polyder(np.poly1d(roughpars)).coefficients
        der = [pgf.poly(Ei, derco) for Ei in got_peaks_keV]
        range_uncal = [
            (r[0] / d, r[1] / d) if isinstance(r, tuple) else r / d
            for r, d in zip(range_keV, der)
        ]
        n_bins = [
            int(sum(r) / 0.5 / d) if isinstance(r, tuple) else int(r / 0.2 / d)
            for r, d in zip(range_keV, der)
        ]

    (
        pk_pars,
        pk_errors,
        pk_covs,
        pk_binws,
        pk_ranges,
        pk_pvals,
        valid_pks,
        pk_funcs,
    ) = hpge_fit_E_peaks(
        E_uncal,
        got_peaks_locs,
        range_uncal,
        n_bins=n_bins,
        funcs=funcs,
        method=method,
        gof_funcs=gof_funcs,
        n_events=n_events,
        uncal_is_int=False,
        simplex=simplex,
        allowed_p_val=allowed_p_val,
        tail_weight=tail_weight,
    )
    results["pk_pars"] = pk_pars
    results["pk_errors"] = pk_errors
    results["pk_covs"] = pk_covs
    results["pk_binws"] = pk_binws
    results["pk_ranges"] = pk_ranges
    results["pk_pvals"] = pk_pvals
    results["pk_validities"] = valid_pks
    results["pk_funcs"] = pk_funcs
    # Drop failed fits
    fitidx = [i == True for i in valid_pks]
    fitted_peaks_keV = results["fitted_keV"] = got_peaks_keV[fitidx]
    funcs = [f for i, f in zip(fitidx, funcs) if i]
    pk_pars = np.asarray(pk_pars, dtype=object)[fitidx]  # ragged
    pk_errors = np.asarray(pk_errors, dtype=object)[fitidx]
    pk_covs = np.asarray(pk_covs, dtype=object)[fitidx]
    pk_binws = np.asarray(pk_binws)[fitidx]
    pk_ranges = np.asarray(pk_ranges)[fitidx]
    pk_pvals = np.asarray(pk_pvals)[fitidx]
    pk_funcs = np.asarray(pk_funcs)[fitidx]
    log.info(f"{sum(fitidx)} peaks fitted:")
    for i, (Ei, parsi, errorsi, covsi, func_i) in enumerate(
        zip(fitted_peaks_keV, pk_pars, pk_errors, pk_covs, pk_funcs)
    ):
        varnames = func_i.__code__.co_varnames[1 : len(pk_pars[-1]) + 1]
        parsi = np.asarray(parsi, dtype=float)
        errorsi = np.asarray(errorsi, dtype=float)
        covsi = np.asarray(covsi, dtype=float)
        # parsigsi = np.sqrt(covsi.diagonal())
        log.info(f"\tEnergy: {str(Ei)}")
        log.info(f"\t\tParameter  |    Value +/- Sigma  ")
        for vari, pari, errorsi in zip(varnames, parsi, errorsi):
            log.info(
                f'\t\t{str(vari).ljust(10)} | {("%4.2f" % pari).rjust(8)} +/- {("%4.2f" % errorsi).ljust(8)}'
            )
            # fwhm??

    # Do a second calibration to the results of the full peak fits
    mus = [
        pgf.get_mu_func(func_i, pars_i, errors=errors_i)
        for func_i, pars_i, errors_i in zip(pk_funcs, pk_pars, pk_errors)
    ]
    mus, mu_vars = zip(*mus)
    mus = np.asarray(mus)
    mu_vars = np.asarray(mu_vars) ** 2

    try:
        pars, errs, cov = hpge_fit_E_scale(mus, mu_vars, fitted_peaks_keV, deg=deg)
        results["pk_cal_pars"] = pars
        results["pk_cal_errs"] = errs
        results["pk_cal_cov"] = cov
    except ValueError:
        log.error("Failed to fit enough peaks to get accurate calibration")
        return None, None, None, results

    # Invert the E scale fit to get a calibration function
    pars, errs, cov = hpge_fit_E_cal_func(mus, mu_vars, fitted_peaks_keV, pars, deg=deg)

    # Finally, calculate fwhms in keV
    uncal_fwhms = [
        pgf.get_fwhm_func(func_i, pars_i, cov=covs_i)
        for func_i, pars_i, covs_i in zip(pk_funcs, pk_pars, pk_covs)
    ]
    uncal_fwhms, uncal_fwhm_errs = zip(*uncal_fwhms)
    uncal_fwhms = np.asarray(uncal_fwhms)
    uncal_fwhm_errs = np.asarray(uncal_fwhm_errs)
    derco = np.polyder(np.poly1d(pars)).coefficients
    der = [pgf.poly(Ei, derco) for Ei in fitted_peaks_keV]

    cal_fwhms = uncal_fwhms * der
    cal_fwhms_errs = uncal_fwhm_errs * der
    results["pk_fwhms"] = np.asarray(
        [(u * d, e * d) for u, e, d in zip(uncal_fwhms, uncal_fwhm_errs, der)]
    )

    log.info(f"{len(cal_fwhms)} FWHMs found:")
    log.info(f"\t   Energy   | FWHM  ")
    for i, (Ei, fwhm, fwhme) in enumerate(
        zip(fitted_peaks_keV, cal_fwhms, cal_fwhms_errs)
    ):
        log.info(
            f"\t{i}".ljust(4)
            + str(Ei).ljust(9)
            + f"| {fwhm:.2f}+-{fwhme:.2f} keV".ljust(5)
        )

    return pars, cov, results


def poly_match(xx, yy, deg=-1, rtol=1e-5, atol=1e-8):
    """Find the polynomial function best matching pol(xx) = yy

    Finds the poly fit of xx to yy that obtains the most matches between pol(xx)
    and yy in the np.isclose() sense. If multiple fits give the same number of
    matches, the fit with the best gof is used, where gof is computed only among
    the matches.
    Assumes that the relationship between xx and yy is monotonic

    Parameters
    ----------
    xx : array-like
        domain data array. Must be sorted from least to largest. Must satisfy
        len(xx) >= len(yy)
    yy : array-like
        range data array: the values to which pol(xx) will be compared. Must be
        sorted from least to largest. Must satisfy len(yy) > max(2, deg+2)
    deg : int
        degree of the polynomial to be used. If deg = 0, will fit for a simple
        scaling: scale * xx = yy. If deg = -1, fits to a simple shift in the
        data: xx + shift = yy. Otherwise, deg is equivalent to the deg argument
        of np.polyfit()
    rtol : float
        the relative tolerance to be sent to np.isclose()
    atol : float
        the absolute tolerance to be sent to np.isclose(). Has the same units
        as yy.

    Returns
    -------
    pars: None or array of floats
        The parameters of the best fit of poly(xx) = yy.  Follows the convention
        used for the return value "p" of polyfit. Returns None when the inputs
        are bad.
    i_matches : list of int
        list of indices in xx for the matched values in the best match
    """

    # input handling
    xx = np.asarray(xx)
    yy = np.asarray(yy)
    #    if len(xx) <= len(yy):
    #        print(f"poly_match error: len(xx)={len(xx)} <= len(yy)={len(yy)}")
    #        return None, 0
    deg = int(deg)
    if deg < -1:
        log.error(f"poly_match error: got bad deg = {deg}")
        return None, 0
    req_ylen = max(2, deg + 2)
    if len(yy) < req_ylen:
        log.error(
            f"poly_match error: len(yy) must be at least {req_ylen} for deg={deg}, got {len(yy)}"
        )
        return None, 0

    maxoverlap = min(len(xx), len(yy))

    # build ixtup: the indices in xx to compare with the values in yy
    ixtup = np.array(list(range(maxoverlap)))
    iytup = np.array(list(range(maxoverlap)))
    best_ixtup = None
    best_iytup = None
    n_close = 0
    gof = np.inf  # lower is better gof
    while True:
        xx_i = xx[ixtup]
        yy_i = yy[iytup]
        gof_i = np.inf

        # simple shift
        if deg == -1:
            pars_i = np.array([1, (np.sum(yy_i) - np.sum(xx_i)) / len(yy_i)])
            polxx = xx_i + pars_i[1]

        # simple scaling
        elif deg == 0:
            pars_i = np.array([np.sum(yy_i * xx_i) / np.sum(xx_i * xx_i), 0])
            polxx = pars_i[0] * xx_i

        # generic poly of degree >= 1
        else:
            pars_i = np.polyfit(xx_i, yy_i, deg)
            polxx = np.zeros(len(yy_i))
            xxn = np.ones(len(yy_i))
            polxx = pgf.poly(xx_i, pars_i)

        # by here we have the best polxx. Search for matches and store pars_i if
        # its the best so far
        matches = np.isclose(polxx, yy_i, rtol=rtol, atol=atol)
        n_close_i = np.sum(matches)
        if n_close_i >= n_close:
            gof_i = np.sum(np.power(polxx[matches] - yy_i[matches], 2))
            if n_close_i > n_close or (n_close_i == n_close and gof_i < gof):
                n_close = n_close_i
                gof = gof_i
                pars = pars_i
                best_ixtup = np.copy(ixtup)
                best_iytup = np.copy(iytup)

        # increment ixtup
        # first find the index of ixtup that needs to be incremented
        ii = 0
        while ii < len(ixtup) - 1:
            if ixtup[ii] < ixtup[ii + 1] - 1:
                break
            ii += 1

        # quit if ii is the last index of ixtup and it's already maxed out
        if not (ii == len(ixtup) - 1 and ixtup[ii] == len(xx) - 1):
            # otherwise increment ii and reset indices < ii
            ixtup[ii] += 1
            ixtup[0:ii] = list(range(ii))
            continue

        # increment iytup
        # first find the index of iytup that needs to be incremented
        ii = 0
        while ii < len(iytup) - 1:
            if iytup[ii] < iytup[ii + 1] - 1:
                break
            ii += 1

        # quit if ii is the last index of iytup and it's already maxed out
        if not (ii == len(iytup) - 1 and iytup[ii] == len(yy) - 1):
            # otherwise increment ii and reset indices < ii
            iytup[ii] += 1
            iytup[0:ii] = list(range(ii))
            ixtup = np.array(list(range(len(iytup))))  # (reset ix)
            continue

        if n_close == len(iytup):  # found best
            break

        # reduce overlap
        new_len = len(iytup) - 1
        if new_len < req_ylen:
            break
        ixtup = np.array(list(range(new_len)))
        iytup = np.array(list(range(new_len)))

        best_ixtup = None
        best_iytup = None
        n_close = 0
        gof = np.inf

    return pars, best_ixtup, best_iytup


def get_i_local_extrema(data, delta):
    """Get lists of indices of the local maxima and minima of data

    The "local" extrema are those maxima / minima that have heights / depths of
    at least delta.
    Converted from MATLAB script at: http://billauer.co.il/peakdet.html

    Parameters
    ----------
    data : array-like
        the array of data within which extrema will be found
    delta : scalar
        the absolute level by which data must vary (in one direction) about an
        extremum in order for it to be tagged

    Returns
    -------
    imaxes, imins : 2-tuple ( array, array )
        A 2-tuple containing arrays of variable length that hold the indices of
        the identified local maxima (first tuple element) and minima (second
        tuple element)
    """

    # prepare output
    imaxes, imins = [], []

    # sanity checks
    data = np.asarray(data)
    if not np.isscalar(delta):
        log.error("get_i_local_extrema: Input argument delta must be a scalar")
        return np.array(imaxes), np.array(imins)
    if delta <= 0:
        log.error(f"get_i_local_extrema: delta ({delta}) must be positive")
        return np.array(imaxes), np.array(imins)

    # now loop over data
    imax, imin = 0, 0
    find_max = True
    for i in range(len(data)):
        if data[i] > data[imax]:
            imax = i
        if data[i] < data[imin]:
            imin = i

        if find_max:
            # if the sample is less than the current max by more than delta,
            # declare the previous one a maximum, then set this as the new "min"
            if data[i] < data[imax] - delta:
                imaxes.append(imax)
                imin = i
                find_max = False
        else:
            # if the sample is more than the current min by more than delta,
            # declare the previous one a minimum, then set this as the new "max"
            if data[i] > data[imin] + delta:
                imins.append(imin)
                imax = i
                find_max = True

    return np.array(imaxes), np.array(imins)


def get_i_local_maxima(data, delta):
    return get_i_local_extrema(data, delta)[0]


def get_i_local_minima(data, delta):
    return get_i_local_extrema(data, delta)[1]


def get_most_prominent_peaks(
    energySeries, xlo, xhi, xpb, max_num_peaks=np.inf, test=False
):
    """find the most prominent peaks in a spectrum by looking for spikes in
    derivative of spectrum energySeries: array of measured energies
    max_num_peaks = maximum number of most prominent peaks to find return a
    histogram around the most prominent peak in a spectrum of a given
    percentage of width
    """
    nb = int((xhi - xlo) / xpb)
    hist, bin_edges = np.histogram(energySeries, range=(xlo, xhi), bins=nb)
    bin_centers = pgh.get_bin_centers(bin_edges)

    # median filter along the spectrum, do this as a "baseline subtraction"
    hist_med = medfilt(hist, 21)
    hist = hist - hist_med

    # identify peaks with a scipy function (could be improved ...)
    peak_idxs = find_peaks_cwt(hist, np.arange(1, 6, 0.1), min_snr=5)
    peak_energies = bin_centers[peak_idxs]

    # pick the num_peaks most prominent peaks
    if max_num_peaks < len(peak_energies):
        peak_vals = hist[peak_idxs]
        sort_idxs = np.argsort(peak_vals)
        peak_idxs_max = peak_idxs[sort_idxs[-max_num_peaks:]]
        peak_energies = np.sort(bin_centers[peak_idxs_max])

    if test:
        plt.plot(bin_centers, hist, ls="steps", lw=1, c="b")
        for e in peak_energies:
            plt.axvline(e, color="r", lw=1, alpha=0.6)
        plt.xlabel("Energy [uncal]", ha="right", x=1)
        plt.ylabel("Filtered Spectrum", ha="right", y=1)
        plt.tight_layout()
        plt.show()
        exit()

    return peak_energies


def match_peaks(data_pks, cal_pks):
    """
    Match uncalibrated peaks with literature energy values.
    """
    from itertools import combinations

    from scipy.stats import linregress

    n_pks = len(cal_pks) if len(cal_pks) < len(data_pks) else len(data_pks)

    cal_sets = combinations(range(len(cal_pks)), n_pks)
    data_sets = combinations(range(len(data_pks)), n_pks)

    best_err, best_m, best_b = np.inf, None, None
    for i, cal_set in enumerate(cal_sets):
        cal = cal_pks[list(cal_set)]  # lit energies for this set

        for data_set in data_sets:
            data = data_pks[list(data_set)]  # uncal energies for this set

            m, b, _, _, _ = linregress(data, y=cal)
            err = np.sum((cal - (m * data + b)) ** 2)

            if err < best_err:
                best_err, best_m, best_b = err, m, b

    print(i, best_err)
    print("cal:", cal)
    print("data:", data)
    plt.scatter(data, cal, label=f"min.err:{err:.2e}")
    xs = np.linspace(data[0], data[-1], 10)
    plt.plot(
        xs, best_m * xs + best_b, c="r", label=f"y = {best_m:.2f} x + {best_b:.2f}"
    )
    plt.xlabel("Energy (uncal)", ha="right", x=1)
    plt.ylabel("Energy (keV)", ha="right", y=1)
    plt.legend()
    plt.tight_layout()
    plt.show()
    exit()

    return best_m, best_b


def calibrate_tl208(energy_series, cal_peaks=None, plotFigure=None):
    """
    energy_series: array of energies we want to calibrate
    cal_peaks: array of peaks to fit
    1.) we find the 2614 peak by looking for the tallest peak at >0.1 the max adc value
    2.) fit that peak to get a rough guess at a calibration to find other peaks with
    3.) fit each peak in peak_energies
    4.) do a linear fit to the peak centroids to find a calibration
    """

    if cal_peaks is None:
        cal_peaks = np.array(
            [238.632, 510.770, 583.191, 727.330, 860.564, 2614.553]
        )  # get_calibration_energies(peak_energies)
    else:
        cal_peaks = np.array(cal_peaks)

    if len(energy_series) < 100:
        return 1, 0

    # get 10 most prominent ~high e peaks
    max_adc = np.amax(energy_series)
    energy_hi = energy_series  # [ (energy_series > np.percentile(energy_series, 20)) & (energy_series < np.percentile(energy_series, 99.9))]

    peak_energies, peak_e_err = get_most_prominent_peaks(
        energy_hi,
    )
    rough_kev_per_adc, rough_kev_offset = match_peaks(peak_energies, cal_peaks)
    e_cal_rough = rough_kev_per_adc * energy_series + rough_kev_offset

    # return rough_kev_per_adc, rough_kev_offset
    # print(energy_series)
    # plt.ion()
    # plt.figure()
    # # for peak in cal_peaks:
    # #     plt.axvline(peak, c="r", ls=":")
    # # energy_series.hist()
    # # for peak in peak_energies:
    # #      plt.axvline(peak, c="r", ls=":")
    # #
    # plt.hist(energy_series)
    # # plt.hist(e_cal_rough[e_cal_rough>100], bins=2700)
    # val = input("do i exist?")
    # exit()

    ###############################################
    # Do a real fit to every peak in peak_energies
    ###############################################
    max_adc = np.amax(energy_series)

    peak_num = len(cal_peaks)
    centers = np.zeros(peak_num)
    fit_result_map = {}
    bin_size = 0.2  # keV

    if plotFigure is not None:
        plot_map = {}

    for i, energy in enumerate(cal_peaks):
        window_width = 10  # keV
        window_width_in_adc = (window_width) / rough_kev_per_adc
        energy_in_adc = (energy - rough_kev_offset) / rough_kev_per_adc
        bin_size_adc = (bin_size) / rough_kev_per_adc

        peak_vals = energy_series[
            (energy_series > energy_in_adc - window_width_in_adc)
            & (energy_series < energy_in_adc + window_width_in_adc)
        ]

        peak_hist, bins = np.histogram(
            peak_vals,
            bins=np.arange(
                energy_in_adc - window_width_in_adc,
                energy_in_adc + window_width_in_adc + bin_size_adc,
                bin_size_adc,
            ),
        )
        bin_centers = pgh.get_bin_centers(bins)
        # plt.ion()
        # plt.figure()
        # plt.plot(bin_centers,peak_hist,  color="k", ls="steps")

        # inp = input("q to quit...")
        # if inp == "q": exit()

        try:
            guess_e, guess_sigma, guess_area = get_gaussian_guess(
                peak_hist, bin_centers
            )
        except IndexError:
            print(f"\n\nIt looks like there may not be a peak at {energy} keV")
            print("Here is a plot of the area I'm searching for a peak...")
            plt.ion()
            plt.figure(figsize=(12, 6))
            plt.subplot(121)
            plt.plot(bin_centers, peak_hist, color="k", ls="steps")
            plt.subplot(122)
            plt.hist(e_cal_rough, bins=2700, histtype="step")
            input("-->press any key to continue...")
            sys.exit()

        plt.plot(
            bin_centers, gauss(bin_centers, guess_e, guess_sigma, guess_area), color="b"
        )

        # inp = input("q to quit...")
        # if inp == "q": exit()

        bounds = (
            [0.9 * guess_e, 0.5 * guess_sigma, 0, 0, 0, 0, 0],
            [
                1.1 * guess_e,
                2 * guess_sigma,
                0.1,
                0.75,
                window_width_in_adc,
                10,
                5 * guess_area,
            ],
        )
        params = fit_binned(
            radford_peak,
            peak_hist,
            bin_centers,
            [guess_e, guess_sigma, 1e-3, 0.7, 5, 0, guess_area],
        )  # bounds=bounds)

        plt.plot(bin_centers, radford_peak(bin_centers, *params), color="r")

        # inp = input("q to quit...")
        # if inp == "q": exit()

        fit_result_map[energy] = params
        centers[i] = params[0]

        if plotFigure is not None:
            plot_map[energy] = (bin_centers, peak_hist)

    # Do a linear fit to find the calibration
    linear_cal = np.polyfit(centers, cal_peaks, deg=1)

    if plotFigure is not None:
        plt.figure(plotFigure.number)
        plt.clf()

        grid = gs.GridSpec(peak_num, 3)
        ax_line = plt.subplot(grid[:, 1])
        ax_spec = plt.subplot(grid[:, 2])

        for i, energy in enumerate(cal_peaks):
            ax_peak = plt.subplot(grid[i, 0])
            bin_centers, peak_hist = plot_map[energy]
            params = fit_result_map[energy]
            ax_peak.plot(
                bin_centers * rough_kev_per_adc + rough_kev_offset,
                peak_hist,
                ls="steps-mid",
                color="k",
            )
            fit = radford_peak(bin_centers, *params)
            ax_peak.plot(
                bin_centers * rough_kev_per_adc + rough_kev_offset, fit, color="b"
            )

        ax_peak.set_xlabel("Energy [keV]")

        ax_line.scatter(
            centers,
            cal_peaks,
        )

        x = np.arange(0, max_adc, 1)
        ax_line.plot(x, linear_cal[0] * x + linear_cal[1])
        ax_line.set_xlabel("ADC")
        ax_line.set_ylabel("Energy [keV]")

        energies_cal = energy_series * linear_cal[0] + linear_cal[1]
        peak_hist, bins = np.histogram(energies_cal, bins=np.arange(0, 2700))
        ax_spec.semilogy(pgh.get_bin_centers(bins), peak_hist, ls="steps-mid")
        ax_spec.set_xlabel("Energy [keV]")

    return linear_cal


def get_calibration_energies(cal_type):
    if cal_type == "th228":
        return np.array(
            [
                238,
                277,
                300,
                452,
                510.77,
                583.191,
                727,
                763,
                785,
                860.564,
                1620,
                2614.533,
            ],
            dtype="double",
        )

    elif cal_type == "uwmjlab":
        # return np.array([239, 295, 351, 510, 583, 609, 911, 969, 1120,
        #                  1258, 1378, 1401, 1460, 1588, 1764, 2204, 2615],
        #                 dtype="double")
        return np.array([239, 911, 1460, 1764, 2615], dtype="double")
    else:
        raise ValueError



from warnings import warn

from lgdo import *  # noqa: F403, F401

warn(
    "pygama.lgdo has moved to its own package, legend-pydataobj. "
    "Please replace 'import pygama.lgdo' with 'import lgdo'. "
    "pygama.lgdo will be removed in a future release.",
    DeprecationWarning,
    stacklevel=2,
)



"""Utility functions."""

from __future__ import annotations

import logging
import re
from datetime import datetime, timezone

import numpy as np
import pandas as pd
from lgdo import Array, ArrayOfEqualSizedArrays, Table, VectorOfVectors, WaveformTable

log = logging.getLogger(__name__)


def to_datetime(key: str) -> datetime:
    """Convert LEGEND cycle key to :class:`~datetime.datetime`.

    Assumes `key` is formatted as ``YYYYMMDDTHHMMSSZ`` (UTC).
    """
    m = re.match(r"^(\d{4})(\d{2})(\d{2})T(\d{2})(\d{2})(\d{2})Z$", key)
    if m is None:
        raise ValueError(f"Could not parse '{key}' as a datetime object")
    else:
        g = [int(el) for el in m.groups()]
        return datetime(*g, tzinfo=timezone.utc)


def to_unixtime(key: str) -> int:
    """Convert LEGEND cycle key to `POSIX timestamp <https://en.wikipedia.org/wiki/Unix_time>`_."""
    return int(to_datetime(key).timestamp())


def inplace_sort(df: pd.DataFrame, by: str) -> None:
    # sort rows according to timestamps
    log.debug(f"sorting database entries according to {by}")
    if by == "timestamp":
        df["_datetime"] = df["timestamp"].apply(to_datetime)
        df.sort_values("_datetime", ignore_index=True, inplace=True)
        df.drop("_datetime", axis=1, inplace=True)
    else:
        df.sort_values(by, ignore_index=True, inplace=True)


def dict_to_table(col_dict: dict, attr_dict: dict):
    for col in col_dict.keys():
        if isinstance(col_dict[col], list):
            if isinstance(col_dict[col][0], (list, np.ndarray, Array)):
                # Convert to VectorOfVectors if there is array-like in a list
                col_dict[col] = VectorOfVectors(
                    listoflists=col_dict[col], attrs=attr_dict[col]
                )
            else:
                # Elements are scalars, convert to Array
                nda = np.array(col_dict[col])
                col_dict[col] = Array(nda=nda, attrs=attr_dict[col])
        elif isinstance(col_dict[col], dict):
            # Dicts are Tables
            col_dict[col] = dict_to_table(
                col_dict=col_dict[col], attr_dict=attr_dict[col]
            )
        else:
            # ndas are Arrays or AOESA
            nda = np.array(col_dict[col])
            if len(nda.shape) == 2:
                dt = attr_dict[col]["datatype"]
                g = re.match(r"\w+<(\d+),(\d+)>{\w+}", dt).groups()
                dims = [int(e) for e in g]
                col_dict[col] = ArrayOfEqualSizedArrays(
                    dims=dims, nda=nda, attrs=attr_dict[col]
                )
            else:
                col_dict[col] = Array(nda=nda, attrs=attr_dict[col])
        attr_dict.pop(col)
    if set(col_dict.keys()) == {"t0", "dt", "values"}:
        return WaveformTable(
            t0=col_dict["t0"],
            dt=col_dict["dt"],
            values=col_dict["values"],
            attrs=attr_dict,
        )
    else:
        return Table(col_dict=col_dict)


def fill_col_dict(
    tier_table: Table,
    col_dict: dict,
    attr_dict: dict,
    tcm_idx: list | pd.RangeIndex,
    table_length: int,
    aoesa_to_vov: bool,
):
    # Put the information from the tier_table (after the columns have been exploded)
    # into col_dict, which will be turned into the final Table
    for col in tier_table.keys():
        if col not in attr_dict.keys():
            attr_dict[col] = tier_table[col].attrs
        else:
            if attr_dict[col] != tier_table[col].attrs:
                if isinstance(tier_table[col], Table):
                    temp_attr = {
                        k: attr_dict[col][k]
                        for k in attr_dict[col].keys() - tier_table[col].keys()
                    }
                    if temp_attr != tier_table[col].attrs:
                        raise ValueError(
                            f"{col} attributes are inconsistent across data"
                        )
                else:
                    raise ValueError(f"{col} attributes are inconsistent across data")
        if isinstance(tier_table[col], ArrayOfEqualSizedArrays):
            # Allocate memory for column for all channels
            if aoesa_to_vov:  # convert to VectorOfVectors
                if col not in col_dict.keys():
                    col_dict[col] = [[]] * table_length
                for i, idx in enumerate(tcm_idx):
                    col_dict[col][idx] = tier_table[col].nda[i]
            else:  # Try to make AoESA, raise error otherwise
                if col not in col_dict.keys():
                    col_dict[col] = np.empty(
                        (table_length, len(tier_table[col].nda[0])),
                        dtype=tier_table[col].dtype,
                    )
                try:
                    col_dict[col][tcm_idx] = tier_table[col].nda
                except BaseException:
                    raise ValueError(
                        f"self.aoesa_to_vov is False but {col} is a jagged array"
                    )
        elif isinstance(tier_table[col], VectorOfVectors):
            # Allocate memory for column for all channels
            if col not in col_dict.keys():
                col_dict[col] = [[]] * table_length
            for i, idx in enumerate(tcm_idx):
                col_dict[col][idx] = tier_table[col][i]
        elif isinstance(tier_table[col], Array):
            # Allocate memory for column for all channels
            if col not in col_dict.keys():
                col_dict[col] = np.empty(
                    table_length,
                    dtype=tier_table[col].dtype,
                )
            col_dict[col][tcm_idx] = tier_table[col].nda
        elif isinstance(tier_table[col], Table):
            if col not in col_dict.keys():
                col_dict[col] = {}
            col_dict[col], attr_dict[col] = fill_col_dict(
                tier_table[col],
                col_dict[col],
                attr_dict[col],
                tcm_idx,
                table_length,
                aoesa_to_vov,
            )
        else:
            log.warning(
                f"not sure how to handle column {col} "
                f"of type {type(tier_table[col])} yet"
            )
    return col_dict, attr_dict



from warnings import warn

from dspeed import *  # noqa: F403, F401

warn(
    "pygama.dsp has moved to its own package, dspeed. "
    "Please replace 'import pygama.dsp' with 'import dspeed'. "
    "pygama.dsp will be removed in a future release.",
    DeprecationWarning,
    stacklevel=2,
)



from __future__ import annotations

import numpy as np
import pandas as pd

# later on we might want a tcm class, or interface / inherit from an entry list
# class. For now we just need the key clustering functionality


def generate_tcm_cols(
    coin_data: list[np.ndarray],
    coin_window: float = 0,
    window_ref: str = "last",
    array_ids: list[int] = None,
    array_idxs: list[int] = None,
) -> dict[np.ndarray]:
    r"""Generate the columns of a time coincidence map.

    Generate the columns of a time coincidence map from a list of arrays of
    coincidence data (e.g. hit times from different channels). Returns 3
    :class:`numpy.ndarray`\ s representing a vector-of-vector-like structure:
    two flattened arrays ``array_id`` (e.g. channel number) and  ``array_idx``
    (e.g. hit index) that specify the location in the input ``coin_data`` of
    each datum belonging to a coincidence event, and a ``cumulative_length``
    array that specifies which rows of the other two output arrays correspond
    to which coincidence event. These can be used to retrieve other data at
    the same tier as the input data into coincidence structures.

    The 0'th entry of ``cumulative_length`` contains the number of hits in the
    zeroth coincidence event, and the i'th entry is set to
    ``cumulative_length[i-1]`` plus the number of hits in the i'th event.
    Thus, the hits of the i'th event can be found in rows
    ``cumulative_length[i-1]`` to ``cumulative_length[i] - 1`` of ``array_id``
    and ``array_idx``.

    An example: ``cumulative_length = [4, 7, ...]``.  Then rows 0 to 3 in
    `array_id` and `array_idx` correspond to the hits in event 0, rows 4 to 6
    correspond to event 1, and so on.

    Makes use of :func:`pandas.concat`, :meth:`pandas.DataFrame.sort_values`,
    and :meth:`pandas.DataFrame.diff` functions:

    - pull data into a :class:`pandas.DataFrame`
    - sort events by strictly ascending value of `coin_col`
    - group hits if the difference in `coin_data` is less than `coin_window`

    Parameters
    ----------
    coin_data
        a list of arrays of the data to be clustered.
    coin_window
        the clustering window width. `coin_data` within the `coin_window` get
        aggregated into the same coincidence cluster. A value of ``0`` means an
        equality test.
    window_ref
        when testing one datum for inclusion in a cluster, test if it is within
        `coin_window` of

        - ``"first"`` -- the first element in the cluster (rigid window width)
        - ``"last"`` -- the last element in the cluster (window grows until two
          data are separated by more than coin_window)

    array_ids
        if provided, use `array_ids` in place of "index in coin_data" as the
        integer corresponding to each element of `coin_data` (e.g. a channel
        number).
    array_idxs
        if provided, use these values in places of the ``DataFrame`` index for
        the return values of `array_idx`.

    Returns
    -------
    col_dict
        keys are ``cumulative_length``, ``array_id``, and  ``array_idx``.
        ``cumulative_length`` specifies which rows of the other two output
        arrays correspond to which coincidence event. ``array_id`` and
        ``array_idx`` specify the location in ``coin_data`` of each datum
        belonging to the coincidence event.
    """
    dfs = []
    for ii, array in enumerate(coin_data):
        array = np.array(array)
        array_id = array_ids[ii] if array_ids is not None else ii
        array_id = np.full_like(array, array_id, dtype=int)
        col_dict = {"array_id": array_id, "coin_data": array}
        if array_idxs is not None:
            col_dict["array_idx"] = array_idxs.astype(int)[ii]
        dfs.append(pd.DataFrame(col_dict, copy=False))  # don't copy the data!

    # concat and sort
    tcm = pd.concat(dfs).sort_values(["coin_data", "array_id"])

    # compute coin_data diffs
    tcm["dcoin"] = tcm["coin_data"].diff()

    # window into coincidences
    # In the future, can add more options, like mean/median/mode
    if window_ref == "last":
        # create the event column by comparing the time since last event to the coincindence window
        tcm["coin_idx"] = (tcm.dcoin > coin_window).cumsum()
    else:
        raise NotImplementedError(f"window_ref {window_ref}")

    # now build the outputs
    cumulative_length = np.where(tcm.coin_idx.diff().to_numpy() != 0)[0]
    cumulative_length[:-1] = cumulative_length[1:]
    cumulative_length[-1] = len(tcm.coin_idx)
    array_id = tcm.array_id.to_numpy()
    array_idx = (
        tcm.array_idx.to_numpy() if "array_idx" in tcm else tcm.index.to_numpy()
    )  # beautiful!
    return {
        "cumulative_length": cumulative_length,
        "array_id": array_id,
        "array_idx": array_idx,
    }



"""Routines for high-level data loading and skimming."""

from __future__ import annotations

import json
import logging
import os
import re
import string
from itertools import product
from keyword import iskeyword
from typing import Iterator

import numpy as np
import pandas as pd
from dspeed.vis import WaveformBrowser
from lgdo import Array, LH5Iterator, LH5Store, Struct, Table, lgdo_utils
from lgdo.types.vectorofvectors import build_cl, explode_arrays, explode_cl
from tqdm.auto import tqdm

from . import utils
from .file_db import FileDB

log = logging.getLogger(__name__)


class DataLoader:
    """Facilitate loading of processed data across several tiers.

    Where possible, uses a :class:`FileDB` object so that a user can quickly
    select a subset of cycle files for interest, and access information at each
    processing tier.

    Example JSON configuration file:

    .. code-block:: json

        {
            "filedb": "path/to/filedb.h5"
            "levels": {
                "hit": {
                    "tiers": ["raw", "dsp", "hit"]
                },
                "tcm": {
                    "tiers": ["tcm"],
                    "parent": "hit",
                    "child": "evt",
                    "tcm_cols": {
                        "child_idx": "coin_idx",
                        "parent_tb": "array_id",
                        "parent_idx": "array_idx"
                    }
                },
                "evt": {
                    "tiers": ["evt"]
                }
            }
        }


    Examples
    --------

    >>> from pygama.flow import DataLoader
    >>> dl = DataLoader("loader-config.json")
    >>> dl.set_files("file_status == 26 and timestamp == '20220716T130443Z'")
    >>> dl.set_datastreams([3, 6, 8], "ch")
    >>> dl.set_cuts({"hit": "daqenergy > 1000 and AoE > 3", "evt": "muon_veto == False"})
    >>> dl.set_output(fmt="pd.DataFrame", columns=["daqenergy", "channel"])
    >>> data = dl.load()

    Be careful, :meth:`.load()` loads data in memory regardless of its size. If
    loading a lot of data (e.g. waveforms), you might want to do it in chunks.
    :class:`.next()` does exactly this:

    >>> for chunk in dl.load():
    ...   run_my_processing(chunk)

    Advanced Usage:

    >>> from pygama.flow import DataLoader
    >>> dl = DataLoader("loader-config.json", filedb="filedb-config.json")  # or any value accepted by the FileDB constructor
    >>> dl.set_files("all")
    >>> dl.set_datastreams([0], "ch")
    >>> dl.set_cuts({"hit": "wf_max > 30000"})
    >>> el = dl.build_entry_list(tcm_level="tcm", mode="any")
    >>> el.query("hit_table == 20", inplace=True)
    >>> dl.set_output(fmt="pd.DataFrame", columns=["daqenergy", "channel"])
    >>> data = dl.load(el)
    """

    def __init__(
        self,
        config: str | dict,
        filedb: str | dict | FileDB = None,
        file_query: str = None,
    ) -> None:
        """
        Parameters
        ----------
        config
            configuration dictionary or JSON file, see above for a specification.
            Accepts strings in the following format: ::

              path/to/config.json[field1/field2/...]]

            to specify the location of the :class:`DataLoader` configuration in
            the ``config.json`` dictionary, if not at the first level.

        filedb
            the loader needs a file database. It can be specified in multiple ways:

            - an instance of :class:`.FileDB`.
            - an LH5 file containing a :class:`.FileDB` (see also
              :meth:`.FileDB.to_disk`).
            - a :class:`.FileDB` configuration dictionary or JSON file.

            If ``None``, uses the value of the ``filedb`` key in `config` to
            instantiate a :class:`.FileDB` object.

        file_query
            string query that should operate on columns of a :class:`.FileDB`.

        Note
        ----
        No data is loaded in memory at this point.
        """
        # declare all member variables
        self.config: dict = None
        self.filedb: FileDB = None
        self.file_list: list = None
        self.table_list = None
        self.cuts = None
        self.merge_files = True
        self.output_format = "lgdo.Table"
        self.output_columns = None
        self.aoesa_to_vov = False
        self.data = None

        # already set FileDB, if supplied
        if isinstance(filedb, FileDB):
            self.filedb = filedb
        elif filedb is not None:
            self.filedb = FileDB(filedb)

        # load things if available
        if config is not None:
            self.set_config(config)

        # set the file_list
        if file_query is not None:
            self.file_list = list(self.filedb.df.query(file_query).index)

    def set_config(self, config: dict | str) -> None:
        """Load configuration dictionary.

        ``$_`` expands to the config file location, if possible, otherwise the
        current working directory.
        """

        # define directory for $_ path substitution later in the config
        # - if a JSON file is provided, it is set to the directory where the
        #   file is located
        # - if a dict is provided, it is set to the current directory
        config_dir = os.getcwd()
        if isinstance(config, str):
            # the config string supports this syntax:
            #  file.json[level1/level2[/...]]
            # to specify where the data loader config shall be found in the
            # json file
            config_loc = None
            m = re.match(r"^(.*)\[(.*)\]$", config)
            if m is not None:
                g = m.groups()
                config = g[0]
                config_loc = g[1].split("/")

            # if a directory is provided, try looking for a file named
            # config.json
            if os.path.isdir(config):
                config_dir = config
                config = os.path.join(config, "config.json")
            else:
                config_dir = os.path.dirname(config)

            with open(config) as f:
                config = json.load(f)

            if config_loc is not None:
                for loc in config_loc:
                    config = config[loc]

        # look for info in configuration if FileDB is not set
        if self.filedb is None:
            # expand $_ variables
            value = lgdo_utils.expand_vars(
                config["filedb"], substitute={"_": config_dir}
            )
            self.filedb = FileDB(value)

        if not os.path.isdir(self.filedb.data_dir):
            raise FileNotFoundError(
                f"{self.filedb.data_dir} (path to data root directory in FileDB) does not exist"
            )

        self.config = config
        self.data_dir = self.filedb.data_dir
        self.levels = list(config["levels"].keys())
        self.tiers = {}
        self.cut_priority = {}
        self.evts = {}
        self.tcms = {}

        for level in self.levels:
            self.tiers[level] = config["levels"][level]["tiers"]
            # Set cut priority
            if "parent" in config["levels"][level].keys():  # This level is a TCM
                parent = config["levels"][level]["parent"]
                child = config["levels"][level]["child"]
                self.tcms[level] = config["levels"][level]
                self.cut_priority[level] = self.cut_priority[parent] + 2
                self.cut_priority[child] = self.cut_priority[parent] + 1
                self.evts[child] = {"tcm": level, "parent": parent}

                # Set TCM columns to lookup
                if "tcm_cols" not in config["levels"][level].keys():
                    log.warning(
                        f"TCM levels, e.g. {level}, need to specify the TCM lookup columns"
                    )
            else:
                self.cut_priority[level] = 0

    def set_files(self, query: str | list[str], append: bool = False) -> None:
        """Apply a file selection.

        Sets `self.file_list`, which is a list of indices corresponding to the
        rows in the file database.

        Parameters
        ----------
        query
            if single string, defines an operation on the file database columns
            supported by :meth:`pandas.DataFrame.query`. In addition, the
            ``all`` keyword is supported to select all files in the database.
            If list of strings, will be interpreted as key (cycle timestamp) list.
        append
            if ``True``, appends files to the existing `self.file_list`
            instead of overwriting.

        Note
        ----
        Call this function before any other operation. A second call to
        :meth:`set_files` does not replace the current file list, which gets
        instead integrated with the new list. Use :meth:`reset` to reset the
        file query.

        Example
        -------
        >>> dl.set_files("file_status == 26 and timestamp == '20220716T130443Z'")
        """

        if isinstance(query, (list, tuple)) and query:
            query = " or ".join([f"timestamp == '{key}'" for key in query])

        if isinstance(query, str):
            if query.replace(" ", "") == "all":
                inds = list(self.filedb.df.index)
            else:
                inds = list(
                    self.filedb.df.query(query, engine="python", inplace=False).index
                )
        else:
            raise ValueError("bad query format")

        if not inds:
            log.warning("no files matching selection found")

        if append and self.file_list is not None:
            self.file_list += inds
            self.file_list = sorted(list(set(self.file_list)))
        else:
            self.file_list = inds

    def get_file_list(self) -> pd.DataFrame:
        """
        Returns a copy of the file database with its dataframe pared down to
        the current file list.
        """
        return self.filedb.df.iloc[self.file_list]

    def set_datastreams(
        self, ds: list | tuple | np.ndarray, word: str, append: bool = False
    ) -> None:
        """Apply selection on data streams (or channels).

        Sets `self.table_list`.

        Parameters
        ----------
        ds
            identifies the detectors of interest. Can be a list of detector
            names, serial numbers, or channels or a list of subsystems of
            interest e.g.  ``ged``.
        word
            the type of identifier used in ds. Should be a key in the given
            channel map or a word defined in the configuration file.
        append
            if ``True``, appends datastreams to the existing `self.table_list`
            instead of overwriting.

        Example
        -------
        >>> dl.set_datastreams(np.arange(40, 45), "ch")
        """
        if self.table_list is None or not append:
            self.table_list = {}

        ds = list(ds)

        found = False
        for level in self.levels:
            tier = self.tiers[level][0]

            template = self.filedb.table_format[tier]
            fm = string.Formatter()
            parse_arr = np.array(list(fm.parse(template)))
            names = list(parse_arr[:, 1])  # fields required to generate file name
            if len(names) > 0:
                keyword = names[0]

            if word == keyword:
                found = True
                if level in self.table_list.keys():
                    self.table_list[level] += ds
                    self.table_list[level] = sorted(list(set(self.table_list[level])))
                else:
                    self.table_list[level] = ds

        if not found:
            # look for word in channel map
            raise NotImplementedError

    def set_cuts(self, cuts: dict | list, append: bool = False) -> None:
        """Apply a selection on columns in the data tables.

        Parameters
        ----------
        cut
            the cuts on the columns of the data table, e.g. ``trapEftp_cal >
            1000``.  If passing a dictionary, the dictionary should be
            structured as ``dict[tier] = cut_expr``. If passing a list, each
            item in the array should be able to be applied on one level of
            tables. The cuts at different levels will be joined with an AND.
        append
            if True, appends cuts to the existing cuts instead of overwriting

        Example
        -------
        >>> dl.set_cuts({"raw": "daqenergy > 1000", "hit": "AoE > 3"})
        """
        if self.cuts is None or not append:
            self.cuts = {}
        if isinstance(cuts, dict):
            # verify the correct structure
            for key, value in cuts.items():
                if not (key in self.levels and isinstance(value, str)):
                    raise ValueError(
                        r"cuts dictionary must be in the format \{ level: string \}"
                    )
                if key in self.cuts.keys() and append:
                    self.cuts[key] += " and " + value
                else:
                    self.cuts[key] = value
        elif isinstance(cuts, list):
            raise NotImplementedError
            self.cuts = {}
            # TODO Parse strings to match column names so you don't have to specify which level it is

    def set_output(
        self,
        fmt: str = None,
        merge_files: bool = None,
        columns: list = None,
        aoesa_to_vov: bool = None,
    ) -> None:
        """
        Set the parameters for the output format of load

        Parameters
        ---------
        fmt
            ``lgdo.Table`` or ``pd.DataFrame``.
        merge_files
            if ``True``, information from multiple files will be merged into
            one table.
        columns
            the columns that should be copied into the output.
        aoesa_to_vov
            output :class:`.ArrayOfEqualSizedArrays` as :class:`.VectorOfVectors`.

        Example
        -------
        >>> dl.set_output(
        ...   fmt="pd.DataFrame",
        ...   merge_files=False,
        ...   columns=["daqenergy", "trapEmax", "channel"]
        ... )
        """
        if fmt not in ["lgdo.Table", "pd.DataFrame", None]:
            raise ValueError(f"'{fmt}' output format not supported")

        if fmt is not None:
            self.output_format = fmt
        if merge_files is not None:
            self.merge_files = merge_files
        if columns is not None:
            self.output_columns = columns
        if aoesa_to_vov is not None:
            self.aoesa_to_vov = aoesa_to_vov

    def reset(self):
        """Resets all fields to their default values.

        As if this is a newly created data loader.
        """
        self.file_list = None
        self.table_list = None
        self.cuts = None
        self.merge_files = True
        self.output_format = "lgdo.Table"
        self.output_columns = None
        self.aoesa_to_vov = False
        self.data = None

    def build_entry_list(
        self,
        tcm_level: str = None,
        tcm_table: int | str = None,
        mode: str = "only",
        save_output_columns: bool = False,
        in_memory: bool = True,
        output_file: str = None,
    ) -> dict[int, pd.DataFrame] | pd.DataFrame | None:
        """Applies cuts to the tables and files of interest.

        Can only load up to two levels, those joined by `tcm_level`.

        Parameters
        ----------
        tcm_level
            the type of TCM to be used. If ``None``, will only return
            information from lowest level.
        tcm_table
            the identifier of the table inside this TCM level that you want to
            use.  If unspecified, there must only be one table inside a TCM
            file in `tcm_level`.
        mode
            if ``any``, returns every hit in the event if any hit in the event
            passes the cuts. If ``only``, only returns hits that pass the cuts.
        save_output_columns
            if ``True``, saves any columns needed for both the cut and the
            output to the `self.entry_list`.
        in_memory
            if ``True``, returns the generated entry list in memory.
        output_file
            HDF5 file name to write the entry list to.

        Returns
        -------
        entries
            the entry list containing columns for ``{parent}_idx``,
            ``{parent}_table``, ``{child}_idx`` and output columns if
            applicable.  Only returned if `in_memory` is ``True``.

        Note
        ----
        Does *not* load the column information into memory. This is done by
        :meth:`.load`.
        """
        if self.file_list is None:
            raise ValueError("You need to make a query on filedb, use set_files()")

        if not in_memory and output_file is None:
            raise ValueError("If in_memory is False, need to specify an output file")

        if in_memory:
            entries = {}

        log.debug("generating entry list determined by cuts")

        # Default columns in the entry_list
        if tcm_level is None:
            return self.build_hit_entries(save_output_columns, in_memory, output_file)

        # Assumes only one tier in a TCM level
        parent = self.tcms[tcm_level]["parent"]
        child = self.tcms[tcm_level]["child"]
        entry_cols = [
            f"{parent}_table",
            f"{parent}_idx",
            f"{child}_idx",
        ]
        if save_output_columns:
            for_output = []

        # Find out which columns are needed for any cuts
        cut_cols = {}
        # ... and pre-load which tiers need to be loaded to make the cuts
        col_tiers_dict = {}

        for level in [child, parent]:
            cut_cols[level] = []
            if self.cuts is not None and level in self.cuts.keys():
                cut = self.cuts[level]
            else:
                cut = ""
            # String parsing to determine which columns need to be loaded
            # Matches any valid python variable name
            terms = re.findall(r"[^\W0-9]\w*", cut)
            for term in terms:
                # Assumes that column names are valid python variable names
                if term.isidentifier() and not iskeyword(term):
                    cut_cols[level].append(term)
                    # Add column to entry_cols if they are needed for both the cut and the output
                    if self.output_columns is not None:
                        if (
                            term in self.output_columns
                            and term not in entry_cols
                            and save_output_columns
                        ):
                            for_output.append(term)
            col_tiers_dict[level] = self.get_tiers_for_col(
                cut_cols[level], merge_files=False
            )

        if save_output_columns:
            entry_cols += for_output

        sto = LH5Store()

        if log.getEffectiveLevel() >= logging.INFO:
            progress_bar = tqdm(
                desc="Building entry list",
                total=len(self.file_list),
                delay=2,
                unit="keys",
            )

        # Make the entry list for each file
        for file in self.file_list:
            if log.getEffectiveLevel() >= logging.INFO:
                progress_bar.update()
                progress_bar.set_postfix(
                    key=self.filedb.df.iloc[file][self.filedb.sortby]
                )

            # Get the TCM specified
            # Assumes that each TCM level only has one tier
            tcm_tier = self.tiers[tcm_level][0]
            tcm_tables = self.filedb.df.loc[file, f"{tcm_tier}_tables"]
            if len(tcm_tables) > 1 and tcm_table is None:
                raise ValueError(
                    f"There are {len(tcm_tables)} TCM tables, need to specify which to use"
                )
            else:
                if tcm_table is not None:
                    if tcm_table in tcm_tables:
                        tcm_tb = tcm_table
                    else:
                        raise ValueError(
                            f"Table {tcm_table} doesn't exist in {tcm_level}"
                        )
                else:
                    tcm_tb = tcm_tables[0]

            tcm_path = os.path.join(
                self.data_dir,
                self.filedb.tier_dirs[tcm_tier].lstrip("/"),
                self.filedb.df.iloc[file][f"{tcm_tier}_file"].lstrip("/"),
            )

            if not os.path.exists(tcm_path):
                raise FileNotFoundError(f"Can't find TCM file for {tcm_level}")

            tcm_table_name = self.filedb.get_table_name(tcm_tier, tcm_tb)
            try:
                tcm_lgdo, _ = sto.read_object(tcm_table_name, tcm_path)
            except KeyError:
                log.warning(f"Cannot find table {tcm_table_name} in file {tcm_path}")
                continue
            # Have to do some hacky stuff until I get a get_dataframe() method
            tcm_lgdo[self.tcms[tcm_level]["tcm_cols"]["child_idx"]] = Array(
                nda=explode_cl(tcm_lgdo["cumulative_length"].nda)
            )
            tcm_lgdo.pop("cumulative_length")
            tcm_tb = Table(col_dict=tcm_lgdo)
            f_entries = tcm_tb.get_dataframe()
            renaming = {
                self.tcms[tcm_level]["tcm_cols"]["child_idx"]: f"{child}_idx",
                self.tcms[tcm_level]["tcm_cols"]["parent_tb"]: f"{parent}_table",
                self.tcms[tcm_level]["tcm_cols"]["parent_idx"]: f"{parent}_idx",
            }
            f_entries.rename(columns=renaming, inplace=True)
            if self.merge_files:
                f_entries["file"] = file
            # At this point, should have a list of all available hits/evts joined by tcm

            if mode == "any":
                drop_idx = None

            # Perform cuts specified for child or parent level, in that order
            for level in [child, parent]:
                # Extract the cut condition if given, otherwise set to ""
                cut = ""
                if self.cuts is not None:
                    if level in self.cuts.keys():
                        cut = self.cuts[level]

                col_tiers = col_tiers_dict[level]

                # Tables in first tier of event should be the same for all tiers in one level
                tables = self.filedb.df.loc[file, f"{self.tiers[level][0]}_tables"]
                if self.table_list is not None:
                    if level in self.table_list.keys():
                        tables = self.table_list[level]
                    else:
                        continue
                if tables is None:
                    continue

                # Cut any rows of TCM not relating to requested tables
                if level == parent:
                    f_entries.query(f"{level}_table in {tables}", inplace=True)

                for tb in tables:
                    tb_table = None
                    if level == parent:
                        tcm_idx = f_entries.query(f"{level}_table == {tb}").index
                    else:
                        tcm_idx = f_entries.index
                    idx_mask = f_entries.loc[tcm_idx, f"{level}_idx"]
                    for tier in self.tiers[level]:
                        tier_path = os.path.join(
                            self.data_dir,
                            self.filedb.tier_dirs[tier].lstrip("/"),
                            self.filedb.df.loc[file, f"{tier}_file"].lstrip("/"),
                        )
                        if tier in col_tiers[file]["tables"].keys():
                            if tb in col_tiers[file]["tables"][tier]:
                                table_name = self.filedb.get_table_name(tier, tb)
                                try:
                                    tier_table, _ = sto.read_object(
                                        table_name,
                                        tier_path,
                                        field_mask=cut_cols[level],
                                        idx=idx_mask.tolist(),
                                    )
                                except KeyError:
                                    log.warning(
                                        f"Cannot find {table_name} in file {tier_path}"
                                    )
                                    continue
                                if tb_table is None:
                                    tb_table = tier_table
                                else:
                                    tb_table.join(tier_table)
                    if tb_table is None:
                        continue
                    tb_df = tb_table.get_dataframe()
                    tb_df.query(cut, inplace=True)
                    idx_match = f_entries.query(f"{level}_idx in {list(tb_df.index)}")
                    if level == parent:
                        idx_match = idx_match.query(f"{level}_table == {tb}")
                    if mode == "only":
                        keep_idx = idx_match.index
                        drop_idx = set.symmetric_difference(
                            set(tcm_idx), list(keep_idx)
                        )
                        f_entries.drop(drop_idx, inplace=True)
                    elif mode == "any":
                        evts = list(idx_match[f"{child}_idx"].unique())
                        keep_idx = f_entries.query(f"{child}_idx in {evts}").index
                        drop = set.symmetric_difference(
                            set(f_entries.index), list(keep_idx)
                        )
                        if drop_idx is None:
                            drop_idx = drop
                        else:
                            drop_idx = set.intersection(drop_idx, drop)
                    else:
                        raise ValueError("mode must be either 'any' or 'only'")

                    if save_output_columns:
                        for col in tb_df.columns:
                            if col in for_output:
                                f_entries.loc[keep_idx, col] = tb_df[col].tolist()

            if mode == "any":
                if drop_idx is not None:
                    f_entries.drop(index=drop_idx, inplace=True)

            f_entries.reset_index(inplace=True, drop=True)

            if in_memory:
                entries[file] = f_entries
            if output_file:
                # Convert f_entries DataFrame to Struct
                f_dict = f_entries.to_dict("list")
                f_struct = Struct(f_dict)
                if self.merge_files:
                    sto.write_object(f_struct, "entries", output_file, wo_mode="a")
                else:
                    sto.write_object(
                        f_struct, f"entries/{file}", output_file, wo_mode="a"
                    )

        if log.getEffectiveLevel() >= logging.INFO:
            progress_bar.close()

        if in_memory:
            if self.merge_files:
                entries = pd.concat(entries.values(), ignore_index=True)
            return entries

    def build_hit_entries(
        self,
        save_output_columns: bool = False,
        in_memory: bool = True,
        output_file: str = None,
    ) -> dict[int, pd.DataFrame] | pd.DataFrame | None:
        """Called by :meth:`.build_entry_list` to handle the case when
        `tcm_level` is unspecified.

        Ignores any cuts set on levels above lowest level.

        Parameters
        ----------
        save_output_columns
            If ``True``, saves any columns needed for both the cut and the
            output to the entry list.
        in_memory
            If ``True``, returns the generated entry list in memory.
        output_file
            HDF5 file name to write the entry list to.

        Returns
        -------
        entries
            the entry list containing columns for ``{low_level}_idx``,
            ``{low_level}_table``, and output columns if
            applicable.  Only returned if `in_memory` is ``True``.
        """
        low_level = self.levels[0]
        if in_memory:
            entries = {}
        entry_cols = [f"{low_level}_table", f"{low_level}_idx"]

        log.debug(
            f"generating (hit-oriented) entry list corresponding to cuts on the {low_level} level"
        )

        # find out which columns are needed for the cut
        cut_cols = []
        cut = ""
        if self.cuts is not None and low_level in self.cuts.keys():
            cut = self.cuts[low_level]
            # do cut expression string parsing to determine which columns need to be loaded
            # this matches any valid python variable name
            terms = re.findall(r"[^\W0-9]\w*", cut)
            for term in terms:
                if not iskeyword(term):
                    cut_cols.append(term)
                    # Add column to entry_cols if they are needed for both the cut and the output
                    if self.output_columns is not None:
                        if (
                            term in self.output_columns
                            and term not in entry_cols
                            and save_output_columns
                        ):
                            entry_cols.append(term)

        log.debug(f"need to load {cut_cols} columns for applying cuts")
        col_tiers = self.get_tiers_for_col(cut_cols, merge_files=False)

        sto = LH5Store()

        if log.getEffectiveLevel() >= logging.INFO:
            progress_bar = tqdm(
                desc="Building entry list",
                total=len(self.file_list),
                delay=2,
                unit="keys",
            )

        # now we loop over the files in our list
        for file in self.file_list:
            if log.getEffectiveLevel() >= logging.INFO:
                progress_bar.update()
                progress_bar.set_postfix(
                    key=self.filedb.df.iloc[file][self.filedb.sortby]
                )

            log.debug(
                f"building entry list for cycle {self.filedb.df.iloc[file][self.filedb.sortby]}"
            )

            # this dataframe will be associated with the file and will contain
            # the columns needed for the cut as well as the columns requested
            # for the output
            f_entries = pd.DataFrame(columns=entry_cols)

            # check if we were asked to consider only certain data streams
            if self.table_list is not None:
                if low_level in self.table_list.keys():
                    tables = self.table_list[low_level]
            else:
                tables = self.filedb.df.loc[file, f"{self.tiers[low_level][0]}_tables"]

            # now loop over these data streams
            for tb in tables:
                tb_table = None

                # no cut is set, let's include all indices for the data stream
                if cut == "":
                    tier = self.tiers[low_level][0]
                    # reconstruct absolute path to tier file
                    tier_path = os.path.join(
                        self.data_dir,
                        self.filedb.tier_dirs[tier].lstrip("/"),
                        self.filedb.df.iloc[file][f"{tier}_file"].lstrip("/"),
                    )
                    # now read how many rows are there in the file
                    table_name = self.filedb.get_table_name(tier, tb)
                    try:
                        n_rows = sto.read_n_rows(table_name, tier_path)
                    except KeyError:
                        log.warning(f"Cannot find {table_name} in file {tier_path}")
                        continue
                    tb_df = pd.DataFrame(
                        {
                            f"{low_level}_idx": np.arange(n_rows),
                            f"{low_level}_table": tb,
                        }
                    )
                    if self.merge_files:
                        tb_df["file"] = file
                else:
                    # loop over tiers in the lowest level
                    for tier in self.tiers[low_level]:
                        # is the tier involved, considered the columns on which cuts are applied?
                        if (
                            tier in col_tiers[file]["tables"].keys()
                            and tb in col_tiers[file]["tables"][tier]
                        ):
                            # reconstruct absolute path to tier file
                            tier_path = os.path.join(
                                self.data_dir,
                                self.filedb.tier_dirs[tier].lstrip("/"),
                                self.filedb.df.iloc[file][f"{tier}_file"].lstrip("/"),
                            )

                            # load the data from the tier file, just the columns needed for the cut
                            table_name = self.filedb.get_table_name(tier, tb)
                            try:
                                tier_tb, _ = sto.read_object(
                                    table_name, tier_path, field_mask=cut_cols
                                )
                            except KeyError:
                                log.warning(
                                    f"Cannot find {table_name} in file {tier_path}"
                                )
                                continue
                            # join eveything in one table
                            if tb_table is None:
                                tb_table = tier_tb
                            else:
                                tb_table.join(tier_tb)

                    if tb_table is None:
                        continue

                    # convert to DataFrame and apply cuts
                    tb_df = tb_table.get_dataframe()
                    tb_df.query(cut, inplace=True)
                    tb_df[f"{low_level}_table"] = tb
                    tb_df[f"{low_level}_idx"] = tb_df.index

                # final DataFrame
                if f_entries.empty:
                    f_entries = tb_df
                else:
                    f_entries = pd.concat((f_entries, tb_df), ignore_index=True)[
                        entry_cols
                    ]

            if self.merge_files:
                f_entries["file"] = file
            if in_memory:
                entries[file] = f_entries
            if output_file:
                # Convert f_entries DataFrame to Struct
                f_dict = f_entries.to_dict("list")
                f_struct = Struct(f_dict)
                if self.merge_files:
                    sto.write_object(f_struct, "entries", output_file, wo_mode="a")
                else:
                    sto.write_object(
                        f_struct, f"entries/{file}", output_file, wo_mode="a"
                    )

        if log.getEffectiveLevel() >= logging.INFO:
            progress_bar.close()

        if in_memory:
            if self.merge_files:
                entries = pd.concat(entries.values(), ignore_index=True)
            return entries

    def next(
        self, entry_list: pd.DataFrame = None, chunk_size: int = 10000, **kwargs
    ) -> Iterator[Table | Struct | pd.DataFrame]:
        """Loads the requested data from disk in chunks.

        This method should be used instead of :meth:`.load` to handle large
        data sets.

        Note
        ----
        It is a user responsibility to optimize the chunk size in order to
        achieve best performance.

        Parameters
        ----------
        chunk_size
            number of entries to load at each iteration. Adapt based on the
            size of each entry and the amount of memory available on the
            system.
        entry_list, **kwargs
            keyword argument forwarded to :meth:`.load`.

        Returns
        -------
        data
            see :meth:`.load`.

        Examples
        --------
        >>> for chunk in dl.next():
        >>>    # 'chunk' has the same type of the output of dl.load()

        See Also
        --------
        .load
        """
        if entry_list is None:
            entry_list = self.build_entry_list(
                tcm_level=kwargs.get("tcm_level", None), save_output_columns=True
            )

        start = stop = 0
        etot = len(entry_list)
        while stop < etot:
            start = stop

            if stop + chunk_size > etot:
                stop = etot
            else:
                stop += chunk_size

            yield self.load(
                entry_list=entry_list[start:stop].reset_index(drop=True), **kwargs
            )

    def load(
        self,
        entry_list: pd.DataFrame = None,
        in_memory: bool = True,
        output_file: str = None,
        orientation: str = "hit",
        tcm_level: str = None,
    ) -> None | Table | Struct | pd.DataFrame:
        """Loads the requested data from disk.

        Loads the requested columns in `self.output_columns` for the entries in
        the given `entry_list`.

        Parameters
        ----------
        entry_list
            the output of :meth:`.build_entry_list`. If ``None``, builds it
            according to the current configuration.
        in_memory
            if ``True``, returns the loaded data in memory and stores in
            `self.data`.
        output_file
            if not ``None``, writes the loaded data to the specified file.
        orientation
            specifies the orientation of the output table. Can be ``hit`` or
            ``evt``.
        tcm_level
            which TCM was used to create the ``entry_list``.

        Returns
        -------
        data
            The data loaded from disk, as specified by `self.output_format`,
            `self.output_columns`, and `self.merge_files`. Only returned if
            `in_memory` is ``True``.
        """
        # set save_output_columns=True to avoid wasting time
        if entry_list is None:
            entry_list = self.build_entry_list(
                tcm_level=tcm_level, save_output_columns=True
            )

        if not in_memory and output_file is None:
            raise ValueError("if in_memory is False, need to specify an output file")

        if self.output_columns is None or not self.output_columns:
            raise ValueError("need to set output columns to load data")

        if orientation == "hit":
            self.data = self.load_hits(entry_list, in_memory, output_file, tcm_level)
        elif orientation == "evt":
            if tcm_level is None:
                if len(self.tcms) == 1:
                    tcm_level = list(self.tcms.keys())[0]
                else:
                    raise ValueError(
                        "need to specify which coincidence map to use to return event-oriented data"
                    )
            self.data = self.load_evts(entry_list, in_memory, output_file, tcm_level)
        else:
            raise ValueError(
                f"I don't understand what orientation={orientation} means!"
            )

        return self.data

    def load_hits(
        self,
        entry_list: pd.DataFrame,
        in_memory: bool = False,
        output_file: str = None,
        tcm_level: str = None,
    ) -> None | Table | Struct | pd.DataFrame:
        """Called by :meth:`.load` when orientation is ``hit``."""
        if tcm_level is None:
            parent = self.levels[0]
            child = None
            load_levels = [parent]
        else:
            parent = self.tcms[tcm_level]["parent"]
            child = self.tcms[tcm_level]["child"]
            load_levels = [parent, child]

        def explode_evt_cols(el: pd.DataFrame, tier_table: Table):
            # Explode columns from "evt"-style levels, untested
            # Will only work if column has an "nda" attribute
            cum_length = build_cl(el[f"{child}_idx"])
            exp_cols = explode_arrays(
                cum_length,
                [a.nda for a in tier_table.values()],
            )
            tier_table.update(zip(tier_table.keys(), exp_cols))
            return tier_table

        sto = LH5Store()

        if self.merge_files:
            tables = entry_list[f"{parent}_table"].unique()
            field_mask = []
            for col in self.output_columns:
                if col not in entry_list.columns:
                    field_mask.append(col)

            col_tiers = self.get_tiers_for_col(field_mask)
            col_dict = entry_list.to_dict("list")
            attr_dict = {}
            for key in col_dict:
                attr_dict[key] = None
            table_length = len(entry_list)

            tot_iter = 0
            for _, level in product(tables, load_levels):
                for _ in self.tiers[level]:
                    tot_iter += 1

            if log.getEffectiveLevel() >= logging.INFO:
                progress_bar = tqdm(
                    desc="Loading data",
                    total=tot_iter,
                    delay=2,
                )

            for tb, level in product(tables, load_levels):
                gb = entry_list.query(f"{parent}_table == {tb}").groupby("file")
                files = list(gb.groups.keys())
                el_idx = list(gb.groups.values())
                idx_mask = [list(entry_list.loc[i, f"{level}_idx"]) for i in el_idx]

                for tier in self.tiers[level]:
                    if log.getEffectiveLevel() >= logging.INFO:
                        progress_bar.update()
                        progress_bar.set_postfix(table=tb, tier=tier)

                    if tb not in col_tiers[tier]:
                        continue

                    tb_name = self.filedb.get_table_name(tier, tb)
                    tier_paths = [
                        os.path.join(
                            self.data_dir,
                            self.filedb.tier_dirs[tier].lstrip("/"),
                            self.filedb.df.iloc[file][f"{tier}_file"].lstrip("/"),
                        )
                        for file in files
                    ]

                    tier_table, _ = sto.read_object(
                        name=tb_name,
                        lh5_file=tier_paths,
                        idx=idx_mask,
                        field_mask=field_mask,
                    )

                    if level == child:
                        explode_evt_cols(entry_list, tier_table)

                    col_dict, attr_dict = utils.fill_col_dict(
                        tier_table,
                        col_dict,
                        attr_dict,
                        [idx for idx_list in el_idx for idx in idx_list],
                        table_length,
                        self.aoesa_to_vov,
                    )

            if log.getEffectiveLevel() >= logging.INFO:
                progress_bar.close()

            # Convert col_dict to lgdo.Table
            f_table = utils.dict_to_table(col_dict=col_dict, attr_dict=attr_dict)

            if output_file:
                sto.write_object(f_table, "merged_data", output_file, wo_mode="o")
            if in_memory:
                if self.output_format == "lgdo.Table":
                    return f_table
                elif self.output_format == "pd.DataFrame":
                    return f_table.get_dataframe()
                else:
                    raise ValueError(
                        f"'{self.output_format}' output format not supported"
                    )
        else:  # not merge_files
            if in_memory:
                load_out = Struct(attrs={"int_keys": True})

            if log.getEffectiveLevel() >= logging.INFO:
                progress_bar = tqdm(
                    desc="Loading data",
                    total=len(entry_list),
                    delay=2,
                    unit="keys",
                )

            # now loop over the output of build_entry_list()
            for file, f_entries in entry_list.items():
                if log.getEffectiveLevel() >= logging.INFO:
                    progress_bar.update()
                    progress_bar.set_postfix(
                        key=self.filedb.df.iloc[file][self.filedb.sortby]
                    )

                log.debug(
                    f"loading data for cycle key {self.filedb.df.iloc[file][self.filedb.sortby]}"
                )

                field_mask = []

                for col in self.output_columns:
                    if col not in f_entries.columns:
                        field_mask.append(col)

                col_tiers = self.get_tiers_for_col(field_mask)
                col_dict = f_entries.to_dict("list")
                attr_dict = {}
                for key in col_dict:
                    attr_dict[key] = None
                table_length = len(f_entries)

                log.debug(f"will load new columns {field_mask}")

                # loop through each table in entry list and
                # loop through each level we're asked to load from
                for tb, level in product(
                    f_entries[f"{parent}_table"].unique(), load_levels
                ):
                    tcm_idx = f_entries.query(f"{parent}_table == {tb}").index
                    idx_mask = f_entries.loc[tcm_idx, f"{level}_idx"].tolist()

                    # loop over tiers in the level
                    for tier in self.tiers[level]:
                        if tb not in col_tiers[file]["tables"][tier]:
                            continue

                        log.debug(
                            f"...for stream '{self.filedb.get_table_name(tier, tb)}' (at {level} level)"
                        )

                        # path to tier file
                        tier_path = os.path.join(
                            self.data_dir,
                            self.filedb.tier_dirs[tier].lstrip("/"),
                            self.filedb.df.iloc[file][f"{tier}_file"].lstrip("/"),
                        )
                        # this should not happen
                        if not os.path.exists(tier_path):
                            raise FileNotFoundError(tier_path)

                        table_name = self.filedb.get_table_name(tier, tb)
                        tier_table, _ = sto.read_object(
                            table_name,
                            tier_path,
                            idx=idx_mask,
                            field_mask=field_mask,
                        )

                        if level == child:
                            explode_evt_cols(f_entries, tier_table)

                        col_dict, attr_dict = utils.fill_col_dict(
                            tier_table,
                            col_dict,
                            attr_dict,
                            tcm_idx,
                            table_length,
                            self.aoesa_to_vov,
                        )
                        # end tb loop

                # Convert col_dict to lgdo.Table
                f_table = utils.dict_to_table(col_dict, attr_dict)

                if in_memory:
                    load_out.add_field(name=file, obj=f_table)
                if output_file:
                    sto.write_object(f_table, f"{file}", output_file, wo_mode="o")
                # end file loop

            if log.getEffectiveLevel() >= logging.INFO:
                progress_bar.close()
            if load_out:
                load_out.update_datatype()

            if in_memory:
                if self.output_format == "lgdo.Table":
                    return load_out
                elif self.output_format == "pd.DataFrame":
                    for file in load_out.keys():
                        load_out[file] = load_out[file].get_dataframe()
                    return load_out
                else:
                    raise ValueError(
                        f"'{self.output_format}' output format not supported"
                    )

    def load_evts(
        self,
        entry_list: pd.DataFrame = None,
        in_memory: bool = False,
        output_file: str = None,
        tcm_level: str = None,
    ) -> None | Table | Struct | pd.DataFrame:
        """Called by :meth:`load` when orientation is ``evt``."""
        raise NotImplementedError

        parent = self.tcms[tcm_level]["parent"]
        child = self.tcms[tcm_level]["child"]
        load_levels = [parent, child]

        sto = LH5Store()

        if self.merge_files:  # Try to load all information at once
            raise NotImplementedError
        else:  # Not merge_files
            if in_memory:
                load_out = Struct(attrs={"int_keys": True})
            for file, f_entries in entry_list.items():
                field_mask = []
                f_table = None
                # Pre-allocate memory for output columns
                for col in self.output_columns:
                    if col not in f_entries.columns:
                        f_entries[col] = None
                        field_mask.append(col)
                col_tiers = self.get_tiers_for_col(field_mask)
                col_dict = f_entries.to_dict("list")
                for col in col_dict.keys():
                    nda = np.array(col_dict[col])
                    col_dict[col] = Array(nda=nda)

                # Loop through each table in entry list
                for tb in f_entries[f"{parent}_table"].unique():
                    tcm_idx = f_entries.query(f"{parent}_table == {tb}").index
                    for level in load_levels:
                        idx_mask = f_entries.loc[tcm_idx, f"{level}_idx"]

                        for tier in self.tiers[level]:
                            if tb in col_tiers["tables"][tier]:
                                tier_path = os.path.join(
                                    self.data_dir,
                                    self.filedb.tier_dirs[tier].lstrip("/"),
                                    self.filedb.df.iloc[file][f"{tier}_file"].lstrip(
                                        "/"
                                    ),
                                )
                                if os.path.exists(tier_path):
                                    table_name = self.filedb.get_table_name(tier, tb)
                                    tier_table, _ = sto.read_object(
                                        table_name,
                                        tier_path,
                                        idx=idx_mask,
                                        field_mask=field_mask,
                                    )
                                    for col in tier_table.keys():
                                        f_table[col].nda[tcm_idx] = tier_table[
                                            col
                                        ].tolist()
                    # end tb loop
                if in_memory:
                    load_out[file] = f_table
                if output_file:
                    sto.write_object(f_table, f"file{file}", output_file, wo_mode="o")
                # end file loop

            if in_memory:
                if self.output_format == "lgdo.Table":
                    return load_out
                elif self.output_format == "pd.DataFrame":
                    for file in load_out.keys():
                        load_out[file] = load_out[file].get_dataframe()
                    return load_out
                else:
                    raise ValueError(
                        f"'{self.output_format}' output format not supported"
                    )

    def load_iterator(
        self,
        entry_list: pd.DataFrame = None,
        tcm_level: str = None,
        buffer_len: int = 3200,
    ) -> LH5Iterator:
        """Creates an :class:LH5Iterator that will load the requested columns
        in `self.output_columns` for the entries in the given `entry_list` in
        chunks. This is more memory efficient than filling a whole table and
        is recommended for use when loading waveforms.

        Parameters
        ----------
        entry_list
            the output of :meth:`.build_entry_list`. If ``None``, builds it
            according to the current configuration.
        tcm_level
            which TCM was used to create the ``entry_list``.
        buffer_len
            how many entries to load in a single chunk

        Returns
        -------
        data
            LH5 Iterator, which yields (lh5 table, entry, n_entries) when
            iterated over.
        """
        if entry_list is None:
            entry_list = self.build_entry_list(
                tcm_level=tcm_level, save_output_columns=True
            )

        if tcm_level is None:
            parent = self.levels[0]
            child = None
            load_levels = [parent]
        else:
            parent = self.tcms[tcm_level]["parent"]
            child = self.tcms[tcm_level]["child"]
            load_levels = [parent, child]

        if self.merge_files:
            tables = entry_list[f"{parent}_table"].unique()
            field_mask = []
            for col in self.output_columns:
                if col not in entry_list.columns:
                    field_mask.append(col)

            col_tiers = self.get_tiers_for_col(field_mask)

            lh5_it = None
            for level in load_levels:
                for tier in self.tiers[level]:
                    # Build list of files/tables/entries
                    lh5_files = []
                    tb_names = []
                    idx_list = []
                    for tb in tables:
                        if tb not in col_tiers[tier]:
                            continue
                        gb = entry_list.query(f"{parent}_table == {tb}").groupby("file")
                        lh5_files += [
                            os.path.join(
                                self.filedb.tier_dirs[tier].lstrip("/"),
                                self.filedb.df.iloc[file][f"{tier}_file"].lstrip("/"),
                            )
                            for file in gb.groups.keys()
                        ]
                        tb_names += [self.filedb.get_table_name(tier, tb)] * len(gb)
                        idx_list += [
                            list(entry_list.loc[i, f"{level}_idx"])
                            for i in gb.groups.values()
                        ]

                    # Create iterator for this tier and friend to other tiers
                    if len(lh5_files) > 0:
                        lh5_it = LH5Iterator(
                            lh5_files=lh5_files,
                            groups=tb_names,
                            base_path=self.data_dir,
                            entry_list=idx_list,
                            field_mask=field_mask,
                            buffer_len=buffer_len,
                            friend=lh5_it,
                        )

            return lh5_it
        else:  # not merge_files
            raise NotImplementedError

    def load_detector(self, det_id):
        """
        special version of `load` designed to retrieve all file files, tables,
        column names, and potentially calibration/dsp parameters relevant to one
        single detector.
        """
        raise NotImplementedError

    def load_settings(self):
        """
        get metadata stored in raw files, usually from a DAQ machine.
        """
        raise NotImplementedError

    def load_dsp_pars(self, query):
        """
        access the dsp_pars parameter database (probably JSON format) and do
        some kind of query to retrieve parameters of interest for our file list,
        and return some tables.
        """
        raise NotImplementedError

    def load_cal_pars(self, query):
        """
        access the cal_pars parameter database, run a query, and return some tables.
        """
        raise NotImplementedError

    def skim_waveforms(self, mode: str = "hit", hit_list=None, evt_list=None):
        """
        handle this one separately because waveforms can easily fill up memory.
        """
        raise NotImplementedError

    # TODO: automatically get the dsp_config/par_database used for
    #   processing when these are set to None
    def browse(
        self, entry_list: pd.DataFrame = None, buffer_len: int = 128, **kwargs
    ) -> WaveformBrowser:
        """Return a :class:`.WaveformBrowser` object for waveform inspection.

        Parameters
        ----------
        entry_list
            the output of :meth:`.build_entry_list`. If ``None``, builds it
            according to the current configuration.
        buffer_len
            number of waveforms to keep in memory at a time.
        **kwargs
            keyword arguments forwarded to :class:`.WaveformBrowser`.

        See Also
        --------
        .WaveformBrowser
        """
        if entry_list is None:
            entry_list = self.build_entry_list()

        parent = self.levels[0]
        tables = entry_list[f"{parent}_table"].unique()

        lh5_it = None
        for tier in self.tiers[parent]:
            # Build list of files/tables/entries
            lh5_files = []
            tb_names = []
            idx_list = []
            for tb in tables:
                gb = entry_list.query(f"{parent}_table == {tb}").groupby("file")
                lh5_files += [
                    os.path.join(
                        self.filedb.tier_dirs[tier].lstrip("/"),
                        self.filedb.df.iloc[file][f"{tier}_file"].lstrip("/"),
                    )
                    for file in gb.groups.keys()
                ]
                tb_names += [self.filedb.get_table_name(tier, tb)] * len(gb)
                idx_list += [
                    list(entry_list.loc[i, f"{parent}_idx"]) for i in gb.groups.values()
                ]

            # Create iterator for this tier and friend to other tiers
            lh5_it = LH5Iterator(
                lh5_files=lh5_files,
                groups=tb_names,
                base_path=self.data_dir,
                entry_list=idx_list,
                buffer_len=buffer_len,
                field_mask={"tracelist": False},
                friend=lh5_it,
            )

        return WaveformBrowser(lh5_it, buffer_len=buffer_len, **kwargs)

    def get_tiers_for_col(
        self, columns: list | np.ndarray, merge_files: bool = None
    ) -> dict:
        """For each column given, get the tiers and tables in that tier where
        that column can be found.

        Parameters
        ----------
        columns
            the columns to look for.

        Returns
        -------
        col_tiers
            ``col_tiers[file]["tables"][tier]`` gives a list of tables in
            ``tier`` that contain a column of interest.
            ``col_tiers[file]["columns"][column]`` gives the tier that
            ``column`` can be found in. If `self.merge_file`s then `
            col_tiers[tier]`` is a list of tables in ``tier`` that contain a
            column of interest.
        """

        col_tiers = {}

        # filedb.columns is needed, generate it now if not available
        if self.filedb.columns is None:
            self.filedb.scan_tables_columns()

        if merge_files is None:
            merge_files = self.merge_files

        if merge_files:
            for file in self.file_list:
                col_inds = set()
                for i, col_list in enumerate(self.filedb.columns):
                    if not set(col_list).isdisjoint(columns):
                        col_inds.add(i)

                for level in self.levels:
                    for tier in self.tiers[level]:
                        col_tiers[tier] = set()
                        if self.filedb.df.loc[file, f"{tier}_col_idx"] is not None:
                            for i in range(
                                len(self.filedb.df.loc[file, f"{tier}_col_idx"])
                            ):
                                if (
                                    self.filedb.df.loc[file, f"{tier}_col_idx"][i]
                                    in col_inds
                                ):
                                    col_tiers[tier].add(
                                        self.filedb.df.loc[file, f"{tier}_tables"][i]
                                    )
        else:
            # loop over selected files (db entries)
            for file in self.file_list:
                # this is the output object
                col_tiers[file] = {"tables": {}, "columns": {}}
                # Rows of FileDB.columns that include columns that we are interested in
                col_inds = set()
                for i, col_list in enumerate(self.filedb.columns):
                    if not set(list(col_list)).isdisjoint(columns):
                        col_inds.add(i)

                # Loop over tiers
                for level in self.levels:
                    for tier in self.tiers[level]:
                        col_tiers[file]["tables"][tier] = []
                        tier_col_idx = self.filedb.df.loc[file, f"{tier}_col_idx"]
                        if tier_col_idx is not None:
                            # Loop over tables
                            for i in range(len(tier_col_idx)):
                                col_idx = self.filedb.df.loc[file, f"{tier}_col_idx"][i]
                                if col_idx in col_inds:
                                    col_tiers[file]["tables"][tier].append(
                                        self.filedb.df.loc[file, f"{tier}_tables"][i]
                                    )
                                    col_in_tier = set.intersection(
                                        set(self.filedb.columns[col_idx]), set(columns)
                                    )
                                    for c in col_in_tier:
                                        col_tiers[file]["columns"][c] = tier

        return col_tiers

    def __repr__(self) -> str:
        return (
            "DataLoader("
            f"cuts={self.cuts}, "
            f"merge_files={self.merge_files}, "
            f'output_format="{self.output_format}", '
            f"output_columns={self.output_columns}, "
            f"aoesa_to_vov={self.aoesa_to_vov}"
            ")"
        )



import os

import lgdo
import numpy as np
from lgdo import LH5Store

from pygama import evt


def test_generate_tcm_cols(lgnd_test_data):
    f_raw = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_raw.lh5"
    )
    tables = lgdo.ls(f_raw)
    store = LH5Store()
    coin_data = []
    for tbl in tables:
        ts, _ = store.read_object(f"{tbl}/raw/timestamp", f_raw)
        coin_data.append(ts)

    tcm_cols = evt.generate_tcm_cols(
        coin_data, 0, "last", [int(tb[2:]) for tb in tables]
    )
    assert isinstance(tcm_cols, dict)
    for v in tcm_cols.values():
        assert np.issubdtype(v.dtype, np.integer)

    # fmt: off
    assert np.array_equal(
        tcm_cols["cumulative_length"],
        [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
        ],
    )
    assert np.array_equal(
        tcm_cols["array_id"],
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804, 1121600,
            1084804, 1084804, 1084804, 1084803, 1084804, 1084804, 1121600,
            1121600, 1084804, 1121600, 1084804, 1121600, 1084803, 1084803,
            1121600, 1121600, 1121600, 1084803, 1084803, 1084803, 1084803,
            1084803, 1084803,
        ],
    )
    assert np.array_equal(
        tcm_cols["array_idx"],
        [
            0, 0, 0, 1, 1, 2, 2, 3, 4, 5, 1, 6, 7, 3, 4, 8, 5, 9, 6, 2, 3, 7,
            8, 9, 4, 5, 6, 7, 8, 9,
        ],
    )
    # fmt: on


def test_build_tcm(lgnd_test_data, tmptestdir):
    f_raw = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_raw.lh5"
    )
    out_file = f"{tmptestdir}/pygama-test-tcm.lh5"
    evt.build_tcm(
        [(f_raw, ["ch1084803/raw", "ch1084804/raw", "ch1121600/raw"])],
        "timestamp",
        out_file=out_file,
        out_name="hardware_tcm",
        wo_mode="of",
    )
    assert os.path.exists(out_file)
    store = LH5Store()
    obj, n_rows = store.read_object("hardware_tcm", out_file)
    assert isinstance(obj, lgdo.Struct)
    assert list(obj.keys()) == ["cumulative_length", "array_id", "array_idx"]



import os
from pathlib import Path

import lgdo.lh5_store as store
import numpy as np
import pytest
from lgdo import LH5Store, ls

from pygama.hit import build_hit

config_dir = Path(__file__).parent / "configs"


def test_basics(dsp_test_file, tmptestdir):
    outfile = f"{tmptestdir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    build_hit(
        dsp_test_file,
        outfile=outfile,
        hit_config=f"{config_dir}/basic-hit-config.json",
        wo_mode="overwrite",
    )

    assert os.path.exists(outfile)
    assert ls(outfile, "/geds/") == ["geds/hit"]


def test_illegal_arguments(dsp_test_file):
    with pytest.raises(ValueError):
        build_hit(dsp_test_file)

    with pytest.raises(ValueError):
        build_hit(
            dsp_test_file,
            hit_config=f"{config_dir}/basic-hit-config.json",
            lh5_tables_config={},
        )

    with pytest.raises(ValueError):
        build_hit(
            dsp_test_file,
            lh5_tables=[],
            lh5_tables_config={},
        )


def test_lh5_table_configs(dsp_test_file, tmptestdir):
    outfile = f"{tmptestdir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    lh5_tables_config = {"/geds/dsp": f"{config_dir}/basic-hit-config.json"}

    build_hit(
        dsp_test_file,
        outfile=outfile,
        lh5_tables_config=lh5_tables_config,
        wo_mode="overwrite",
    )

    assert os.path.exists(outfile)
    assert ls(outfile, "/geds/") == ["geds/hit"]

    lh5_tables_config = {
        "/geds/dsp": {
            "outputs": ["calE", "AoE"],
            "operations": {
                "calE": {
                    "expression": "sqrt(a + b * trapEmax**2)",
                    "parameters": {"a": 1.23, "b": 42.69},
                },
                "AoE": {"expression": "A_max/calE"},
            },
        }
    }

    build_hit(
        dsp_test_file,
        outfile=outfile,
        lh5_tables_config=lh5_tables_config,
        wo_mode="overwrite",
    )

    assert os.path.exists(outfile)
    assert ls(outfile, "/geds/") == ["geds/hit"]


def test_outputs_specification(dsp_test_file, tmptestdir):
    outfile = f"{tmptestdir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    build_hit(
        dsp_test_file,
        outfile=outfile,
        hit_config=f"{config_dir}/basic-hit-config.json",
        wo_mode="overwrite",
    )

    store = LH5Store()
    obj, _ = store.read_object("/geds/hit", outfile)
    assert list(obj.keys()) == ["calE", "AoE", "A_max"]


def test_build_hit_spms_basic(dsp_test_file_spm, tmptestdir):
    out_file = f"{tmptestdir}/L200-comm-20211130-phy-spms_hit.lh5"
    build_hit(
        dsp_test_file_spm,
        outfile=out_file,
        hit_config=f"{config_dir}/spms-hit-config.json",
        wo_mode="overwrite_file",
    )
    assert ls(out_file) == ["ch0", "ch1", "ch2"]
    assert ls(out_file, "ch0/") == ["ch0/hit"]
    assert ls(out_file, "ch0/hit/") == [
        "ch0/hit/energy_in_pe",
        "ch0/hit/quality_cut",
        "ch0/hit/trigger_pos",
    ]


def test_build_hit_spms_multiconfig(dsp_test_file_spm, tmptestdir):
    out_file = f"{tmptestdir}/L200-comm-20211130-phy-spms_hit.lh5"

    build_hit(
        dsp_test_file_spm,
        outfile=out_file,
        lh5_tables_config=f"{config_dir}/spms-hit-multi-config.json",
        wo_mode="overwrite",
    )
    assert ls(out_file) == ["ch0", "ch1", "ch2"]
    assert ls(out_file, "ch0/") == ["ch0/hit"]
    assert ls(out_file, "ch0/hit/") == [
        "ch0/hit/energy_in_pe",
        "ch0/hit/quality_cut",
        "ch0/hit/trigger_pos",
    ]


def test_build_hit_spms_calc(dsp_test_file_spm, tmptestdir):
    out_file = f"{tmptestdir}/L200-comm-20211130-phy-spms_hit.lh5"

    build_hit(
        dsp_test_file_spm,
        outfile=out_file,
        wo_mode="overwrite_file",
        lh5_tables_config=f"{config_dir}/spms-hit-a-config.json",
    )
    assert ls(out_file) == ["ch0", "ch1", "ch2"]
    assert ls(out_file, "ch0/") == ["ch0/hit"]
    assert ls(out_file, "ch0/hit/") == ["ch0/hit/energy_in_pe"]

    df0 = store.load_nda(out_file, ["energy_in_pe"], "ch0/hit/")
    df1 = store.load_nda(out_file, ["energy_in_pe"], "ch1/hit/")
    df2 = store.load_nda(out_file, ["energy_in_pe"], "ch2/hit/")

    assert len(df0["energy_in_pe"]) == 5
    assert len(df1["energy_in_pe"]) == 5
    assert len(df2["energy_in_pe"]) == 5

    assert len(df0["energy_in_pe"][0]) == 20
    assert len(df1["energy_in_pe"][0]) == 20
    assert len(df2["energy_in_pe"][0]) == 20

    assert np.nanmean(df0["energy_in_pe"]) == 0
    assert np.nanmean(df1["energy_in_pe"]) == 1
    assert np.nanmean(df2["energy_in_pe"]) == 2
